"""Tests for the module graph_pattern."""

import uuid

import networkx as nx
import pytest

from pydexpi.syndata.graph_pattern import (
    GraphBasicPipingConnector,
    GraphBasicSignalConnector,
    GraphPattern,
)


def make_a_graph_pattern(label: str, connector_label_suffix: str = "0"):
    """Make a dummy graph pattern with a single node graph. The pattern has
    four connectors, one of each type.
    """
    single_node_graph = nx.DiGraph()
    the_node = str(str(uuid.uuid4()))
    single_node_graph.add_node(the_node)
    connectors = []
    connectors.append(GraphBasicPipingConnector(f"PipeIn_{connector_label_suffix}", the_node, True))
    connectors.append(
        GraphBasicPipingConnector(f"PipeOut_{connector_label_suffix}", the_node, False)
    )
    connectors.append(
        GraphBasicSignalConnector(f"SignalIn_{connector_label_suffix}", the_node, True)
    )
    connectors.append(
        GraphBasicSignalConnector(f"SignalOut_{connector_label_suffix}", the_node, False)
    )

    return GraphPattern(label, single_node_graph, connectors)


# Test connector
def test_connector_constructor():
    """Test if a graph connector is instantiated correctly."""
    ref_node_id = str(uuid.uuid4())
    new_connector = GraphBasicPipingConnector("Clabel", ref_node_id, True)
    assert new_connector.label == "Clabel"
    assert new_connector.reference_node_id == ref_node_id
    assert new_connector.is_inlet


def test_assess_valid_counterpart():
    """Test the capability to assess a valid counterpart of graph connectors."""
    pinc = GraphBasicPipingConnector("PipeIn", str(uuid.uuid4()), True)
    poutc = GraphBasicPipingConnector("PipeOut", str(uuid.uuid4()), False)
    sinc = GraphBasicSignalConnector("SignalIn", str(uuid.uuid4()), True)
    soutc = GraphBasicSignalConnector("SignalOut", str(uuid.uuid4()), False)

    # Test valid connections
    assert pinc.assess_valid_counterpart(poutc) is True
    assert poutc.assess_valid_counterpart(pinc) is True
    assert sinc.assess_valid_counterpart(soutc) is True
    assert soutc.assess_valid_counterpart(sinc) is True

    # Test invalid connections
    assert pinc.assess_valid_counterpart(pinc) is False
    assert sinc.assess_valid_counterpart(sinc) is False
    assert pinc.assess_valid_counterpart(sinc) is False
    assert soutc.assess_valid_counterpart(poutc) is False


# Pattern tests
def test_pattern_constructor():
    """Test instantiating a pattern"""
    the_pattern = make_a_graph_pattern("Pattern1")
    assert len(the_pattern.the_graph.nodes) == 1
    assert len(the_pattern.connectors) == 4


def test_incorporate_pattern():
    """Test connecting a pattern"""
    pattern1 = make_a_graph_pattern("Pattern1")
    pattern2 = make_a_graph_pattern("Pattern2", connector_label_suffix="1")

    # First, try connecting via invalid connectors
    with pytest.raises(ValueError):
        pattern1.incorporate_pattern(
            list(pattern1.connectors.values())[0], pattern2, list(pattern2.connectors.values())[2]
        )

    # Try to connect via valid connector pair using dictionary keys
    pattern1.incorporate_pattern(
        pattern1.connectors["PipeIn_0"], pattern2, pattern2.connectors["PipeOut_1"]
    )

    assert len(pattern1.the_graph.nodes) == 2
    assert len(pattern1.the_graph.edges) == 1
    assert len(pattern1.connectors) == 6


def test_copy_pattern():
    """Test copying a graph pattern with the overriden copy method."""
    pattern1 = make_a_graph_pattern("Pattern1")
    copied_pattern = pattern1.copy_pattern()
    assert len(copied_pattern.the_graph.nodes) == 1
    assert len(copied_pattern.connectors) == 4
    for connector in copied_pattern.connectors.values():
        assert connector.reference_node_id in copied_pattern.the_graph.nodes


def test_load_and_save(tmp_path):
    """Test loading and saving a graph pattern"""
    the_pattern = make_a_graph_pattern("Pattern1")
    the_pattern.save(tmp_path, "test")
    loaded_pattern = GraphPattern.load(tmp_path, "test")

    assert the_pattern.label == loaded_pattern.label

    connectors = the_pattern.connectors
    loaded_connectors = loaded_pattern.connectors
    assert len(connectors) == len(loaded_connectors)
    for c, lc in zip(connectors, loaded_connectors):
        assert c == lc
        assert connectors[c].label == loaded_connectors[lc].label
        assert connectors[c].reference_node_id == loaded_connectors[lc].reference_node_id

    observers = the_pattern.observer_patterns
    loaded_observers = loaded_pattern.observer_patterns
    assert len(observers) == len(loaded_observers)
    assert set(observers.keys()) == set(loaded_observers.keys())



from .pydantic_classes import (
    Area,
    AreaUnit,
    ElectricalFrequency,
    ElectricalFrequencyUnit,
    Force,
    ForceUnit,
    HeatTransferCoefficient,
    HeatTransferCoefficientUnit,
    Length,
    LengthUnit,
    Mass,
    MassFlowRate,
    MassFlowRateUnit,
    MassUnit,
    NullableArea,
    NullableElectricalFrequency,
    NullableForce,
    NullableFrequency,
    NullableHeatTransferCoefficient,
    NullableLength,
    NullableMass,
    NullableMassFlowRate,
    NullableNumberPerTimeInterval,
    NullablePercentage,
    NullablePower,
    NullablePressure,
    NullablePressureAbsolute,
    NullablePressureGauge,
    NullableRotationalFrequency,
    NullableTemperature,
    NullableVoltage,
    NullableVolume,
    NullableVolumeFlowRate,
    NullArea,
    NullElectricalFrequency,
    NullForce,
    NullHeatTransferCoefficient,
    NullLength,
    NullMass,
    NullMassFlowRate,
    NullNumberPerTimeInterval,
    NullPercentage,
    NullPower,
    NullPressureAbsolute,
    NullPressureGauge,
    NullRotationalFrequency,
    NullTemperature,
    NullVoltage,
    NullVolume,
    NullVolumeFlowRate,
    NumberPerTimeInterval,
    NumberPerTimeIntervalUnit,
    Percentage,
    PercentageUnit,
    Power,
    PowerUnit,
    PressureAbsolute,
    PressureAbsoluteUnit,
    PressureGauge,
    PressureGaugeUnit,
    RotationalFrequency,
    RotationalFrequencyUnit,
    Temperature,
    TemperatureUnit,
    Voltage,
    VoltageUnit,
    Volume,
    VolumeFlowRate,
    VolumeFlowRateUnit,
    VolumeUnit,
)



"""Tests for the GenerationHistory class and related functions."""

from pydexpi.syndata.generation_history import (
    GenerationHistory,
    load_generation_history_from_json,
    save_generation_history_as_json,
)
from pydexpi.syndata.generator_step import GeneratorStepDict


class MockStep:
    """Mock class to simulate a generator step."""

    def __init__(
        self,
        step_type: str,
        own_connector: str = None,
        next_pattern: str = None,
        next_connector: str = None,
        sampled_distribution_name: str = None,
    ):
        """Initialize the mock step."""
        self.step_type = step_type
        self.own_connector = own_connector
        self.next_pattern = next_pattern
        self.next_connector = next_connector
        self.sampled_distribution_name = sampled_distribution_name

    def to_dict(self) -> GeneratorStepDict:
        """Convert the mock step to a dictionary."""
        return {
            "generator_step_type": self.step_type,
            "own_connector": self.own_connector,
            "next_pattern": self.next_pattern,
            "next_connector": self.next_connector,
            "sampled_distribution_name": self.sampled_distribution_name,
        }


def test_generation_history_init():
    """Test that GenerationHistory initializes correctly."""
    history = GenerationHistory()
    assert isinstance(history.history, list)
    assert len(history.history) == 0


def test_write_step():
    """Test writing steps to the history."""
    history = GenerationHistory()

    # Create and add some test steps
    step1 = MockStep("initialization", next_pattern="Pattern1", sampled_distribution_name="dist1")
    step2 = MockStep(
        "add_pattern",
        own_connector="conn1",
        next_pattern="Pattern2",
        next_connector="conn2",
        sampled_distribution_name="dist2",
    )

    history.write_step(step1)
    history.write_step(step2)

    # Check the history contains the expected steps
    assert len(history.history) == 2
    assert history.history[0]["generator_step_type"] == "initialization"
    assert history.history[0]["next_pattern"] == "Pattern1"
    assert history.history[1]["generator_step_type"] == "add_pattern"
    assert history.history[1]["own_connector"] == "conn1"


def test_save_load_generation_history(tmp_path):
    """Test saving and loading generation history."""
    history = GenerationHistory()

    # Create and add some test steps
    step1 = MockStep("initialization", next_pattern="Pattern1", sampled_distribution_name="dist1")
    step2 = MockStep(
        "add_pattern",
        own_connector="conn1",
        next_pattern="Pattern2",
        next_connector="conn2",
        sampled_distribution_name="dist2",
    )

    history.write_step(step1)
    history.write_step(step2)

    # Save the history to a temporary file
    json_path = tmp_path / "test_history.json"
    save_generation_history_as_json(history, json_path)

    # Load the history from the file
    loaded_history = load_generation_history_from_json(json_path)

    # Check that the loaded history matches the original
    assert len(loaded_history) == 2
    assert loaded_history[0]["generator_step_type"] == "initialization"
    assert loaded_history[0]["next_pattern"] == "Pattern1"
    assert loaded_history[1]["generator_step_type"] == "add_pattern"
    assert loaded_history[1]["own_connector"] == "conn1"
    assert loaded_history[1]["next_connector"] == "conn2"



import warnings
from copy import copy

import networkx as nx
import plotly.graph_objects as go
from Flowsheet_Class.utils_visualization import _add_positions
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from networkx import DiGraph

from pydexpi.dexpi_classes import equipment, instrumentation, piping
from pydexpi.dexpi_classes.dexpiBaseModels import DexpiBaseModel
from pydexpi.dexpi_classes.dexpiModel import DexpiModel
from pydexpi.loaders.graph_loader import NXGraphLoader
from pydexpi.toolkits.base_model_utils import (
    get_data_attributes,
    get_dexpi_class,
)


class MLGraphLoader(NXGraphLoader):
    """Loads a DEXPI model instance into networkX graph. The graph is optimized
    for machine learning tasks.

    Attributes
    ----------
        plant_graph (nx.DiGraph): Graph version of a plant.
        plant_model (DexpiModel): Instance of DexpiModel.


    """

    def __init__(
        self,
        plant_graph: DiGraph | None = None,
        plant_model: DexpiModel | None = None,
    ):
        """Initializes a new instance of GraphLoader.

        Args:
        ----
            plant_graph (Optional[DiGraph]): Initialize loader with the graph
                version of a plant. Defaults to an empty graph if not provided.
            plant_model (Optional[PlantModel]): Initialize loader with an
                instance of PlantModel. Defaults to a new instance if not
                provided.

        """
        self.plant_graph = plant_graph if plant_graph is not None else DiGraph()
        self.plant_model = plant_model if plant_model is not None else DexpiModel()

    # Implement abstract methods
    def dexpi_to_graph(self, dexpi_model: DexpiModel) -> DiGraph:
        """
        Parse a pyDEXPI model into a NetworkX graph representation.

        Parameters
        ----------
        pydexpi_model : DexpiModel
            The pyDEXPI model to be parsed into a NetworkX graph.

        Returns
        -------
        nx.Graph
            The parsed NetworkX graph representation of the pyDEXPI model.
        """
        # Reset plant graph and set dexpi model
        self.plant_graph = DiGraph()
        self.plant_model = dexpi_model

        self.parse_equipment_and_piping()
        self.parse_instrumentation()

        return self.plant_graph

    def graph_to_dexpi(self, plant_graph: DiGraph) -> DexpiModel:
        """Parsing the example graph back to pyDEXPI is not supported yet."""
        raise NotImplementedError

    def validate_graph_format(
        self, special_node_attributes: list[str] = [], special_edge_attributes: list[str] = []
    ):
        """Validate a graph against the graph format. Raises an AttributeError if the graph is not valid.

        Parameters
        ----------
        special_node_attributes : list[str]
            Node attributes besides "dexpi_class" and besides attributes of a DEXPI class.
        special_edge_attributes : list[str]
            Edge attributes besides "dexpi_class" and besides attributes of a DEXPI class.
        """
        special_node_attributes = special_node_attributes + ["dexpi_class", "id"]
        special_edge_attributes = special_edge_attributes + ["dexpi_class"]

        for node_attributes in self.plant_graph.nodes.values():
            self.validate_node(node_attributes, special_node_attributes)
        for (source, target), edge_attributes in self.plant_graph.edges.items():
            source_class_name = self.plant_graph.nodes[source]["dexpi_class"]
            target_class_name = self.plant_graph.nodes[target]["dexpi_class"]
            self.validate_edge(
                source_class_name, target_class_name, edge_attributes, special_edge_attributes
            )

    def validate_node(self, node: dict, special_node_attributes: list[str] = []):
        """Validate a node against the DEXPI class and its attributes.

        Parameters
        ----------
        node : dict
            Dictionary describing the node.
        special_node_attributes : list[str]
            Node attributes besides "dexpi_class" and besides attributes of a DEXPI class.
        """
        node_class = get_dexpi_class(node["dexpi_class"])
        self.validate_node_classes(node_class)
        self.validate_node_attributes(node, node_class, special_node_attributes)

    def validate_node_classes(self, node_class: type[DexpiBaseModel]):
        """Validate the node class name against the allowed node classes.
        The following classes are allowed as nodes:
        - NozzleOwner
        - PipingNetworkSegmentItem
        - ProcessInstrumentationFunction
        - SignalOffPageConnector
        - ControlledActuator
        - Positioner
        - ElectronicFrequencyConverter
        - OfflinePrimaryElement

        Parameters
        ----------
        node_class : type[DexpiBaseModel]
            Class to be validated.

        Raises
        ------
        AttributeError
            Node class is not valid.
        """
        node_classes = (
            equipment.NozzleOwner
            | piping.PipingNetworkSegmentItem
            | instrumentation.ProcessInstrumentationFunction
            | instrumentation.SignalOffPageConnector
            | instrumentation.ControlledActuator
            | instrumentation.Positioner
            | instrumentation.ElectronicFrequencyConverter
            | instrumentation.OfflinePrimaryElement
        )
        if not issubclass(node_class, node_classes):
            raise AttributeError(f"'{node_class.__name__}' not a valid DEXPI class or node.")

    def validate_node_attributes(
        self, node: dict, node_class: type[DexpiBaseModel], special_node_attributes: list[str] = []
    ):
        """Validate attributes of a node against the DEXPI class and its allowed attributes.

        Parameters
        ----------
        node : dict
            Attributes of the node.
        node_class : type[DexpiBaseModel]
            Class of the node.
        special_node_attributes : list[str]
            Node attributes besides "dexpi_class" and besides attributes of a DEXPI class.

        Raises
        ------
        AttributeError
            Node attributes are not valid.
        """
        for attribute in node:
            if attribute not in (list(node_class.model_fields) + special_node_attributes):
                raise AttributeError(
                    f"Attribute {attribute} not valid for DEXPI class {node['dexpi_class']}."
                )

    def validate_edge(
        self,
        source_class_name: str,
        target_class_name: str,
        edge: dict,
        special_edge_attributes: list[str] = [],
    ):
        """The following classes are allowed as edges:
        - PipingConnection
        - SignalConveyingFunction
        - OperatedValveReference
        - ElectronicFrequencyConverter


        Parameters
        ----------
        source_class_name : str
            Dexpi class name of the source node.
        target_class_name : str
            Dexpi class name of the target node.
        edge : dict
            Attributes of the edge.
        special_edge_attributes : list[str]
            Edge attributes besides "dexpi_class" and besides attributes of a DEXPI class.
        """
        edge_class = get_dexpi_class(edge["dexpi_class"])
        self.validate_edge_classes(edge_class)
        self.validate_edge_attributes(edge, edge_class, special_edge_attributes)
        source_class = get_dexpi_class(source_class_name)
        self.validate_edge_sources(source_class, edge_class)
        target_class = get_dexpi_class(target_class_name)
        self.validate_edge_targets(target_class, edge_class)

    def validate_edge_classes(self, edge_class: type[DexpiBaseModel]):
        """Validate the edge class name against the allowed edge classes.

        Parameters
        ----------
        edge_class : type[DexpiBaseModel]
            Edge class to be validated.

        Raises
        ------
        AttributeError
            If the edge class is not valid.
        """
        edge_classes = (
            piping.PipingConnection
            | instrumentation.SignalConveyingFunction
            | instrumentation.OperatedValveReference
            | instrumentation.ElectronicFrequencyConverter
        )
        if not issubclass(edge_class, edge_classes):
            raise AttributeError(f"'{edge_class.__name__}' not a valid DEXPI class or edge.")

    def validate_edge_attributes(
        self, edge: dict, edge_class: type[DexpiBaseModel], special_edge_attributes: list[str] = []
    ):
        """Validate the attributes of an edge against the DEXPI class and its allowed attributes.

        Parameters
        ----------
        edge : dict
            Attributes of the edge.
        edge_class : type[DexpiBaseModel]
            Class of the edge.
        special_edge_attributes : list[str]
            Edge attributes besides "dexpi_class" and besides attributes of a DEXPI class.

        Raises
        ------
        AttributeError
            Edge attributes are not valid.
        """
        valid_attributes = copy(edge_class.model_fields)
        # Special handling for PipingConnection
        if issubclass(edge_class, piping.PipingConnection):
            valid_attributes.update(piping.PipingNetworkSegment.model_fields)

        for key in edge.keys():
            if key not in (list(valid_attributes) + special_edge_attributes):
                raise AttributeError(
                    f"Attribute {key} not valid for DEXPI class {edge['dexpi_class']}."
                )

    def validate_edge_sources(
        self, source_class: type[DexpiBaseModel], edge_class: type[DexpiBaseModel]
    ):
        """Validates if the source node is allowed to be connected to the edge.

        Parameters
        ----------
        source_class : type[DexpiBaseModel]
            Source node class.
        edge_class : type[DexpiBaseModel]
            Edge class.

        Raises
        ------
        AttributeError
            Edge source is not valid.
        """
        if issubclass(edge_class, piping.PipingConnection):
            if not issubclass(
                source_class,
                (equipment.NozzleOwner | piping.PipingNetworkSegmentItem),
            ):
                raise AttributeError(
                    f"Source of {edge_class.__name__} must be NozzleOwner or PipingNetworkSegmentItem."
                )

        elif issubclass(edge_class, instrumentation.SignalConveyingFunction):
            if not issubclass(
                source_class,
                (
                    equipment.NozzleOwner
                    | piping.PipingNetworkSegmentItem
                    | instrumentation.ProcessInstrumentationFunction
                    | instrumentation.SignalOffPageConnector
                    | instrumentation.OfflinePrimaryElement
                ),
            ):
                raise AttributeError(
                    f"Source of {edge_class.__name__} should be connected to at least one instrumentation object."
                )

        elif issubclass(edge_class, instrumentation.OperatedValveReference):
            if not issubclass(
                source_class, (instrumentation.ControlledActuator | instrumentation.Positioner)
            ):
                raise AttributeError(
                    f"Source of {edge_class.__name__} should connect ControlledActuator or Positioner to (subclass of) OperatedValve."
                )

        elif issubclass(edge_class, instrumentation.ElectronicFrequencyConverter):
            if not issubclass(source_class, instrumentation.ProcessInstrumentationFunction):
                raise AttributeError(
                    f"Source of {edge_class.__name__} must be ProcessInstrumentationFunction."
                )

    def validate_edge_targets(
        self, target_class: type[DexpiBaseModel], edge_class: type[DexpiBaseModel]
    ):
        """Validates if the target node is allowed to be connected to the edge.

        Parameters
        ----------
        target_class : type[DexpiBaseModel]
            Target node of the edge.
        edge_class : type[DexpiBaseModel]
            Edge class.

        Raises
        ------
        AttributeError
            Edge target is not valid.
        """
        if issubclass(edge_class, piping.PipingConnection):
            if not issubclass(
                target_class,
                (equipment.NozzleOwner | piping.PipingNetworkSegmentItem),
            ):
                raise AttributeError(
                    f"Target of {edge_class.__name__} must be NozzleOwner or PipingNetworkSegmentItem."
                )

        elif issubclass(edge_class, instrumentation.SignalConveyingFunction):
            if not issubclass(
                target_class,
                (
                    instrumentation.ProcessInstrumentationFunction
                    | instrumentation.ControlledActuator
                    | instrumentation.Positioner
                    | equipment.Equipment  # assume that SignalConveyingFunction has target ActuatingElectricalFunction which has a Nozzle of an Equipment as SensingLocation
                ),
            ):
                raise AttributeError(
                    f"Target of {edge_class.__name__} must be ProcessInstrumentationFunction, ControlledActuator, or Positioner."
                )

        elif issubclass(edge_class, instrumentation.OperatedValveReference):
            if not issubclass(target_class, piping.OperatedValve):
                raise AttributeError(
                    f"Target of {edge_class.__name__} must be (subclass of) OperatedValve."
                )

        elif issubclass(edge_class, instrumentation.ElectronicFrequencyConverter):
            if not issubclass(target_class, instrumentation.ElectronicFrequencyConverter):
                raise AttributeError(
                    f"Target of {edge_class.__name__} must be ElectronicFrequencyConverter."
                )

    def add_node(self, obj: DexpiBaseModel) -> None:
        """Adds an object obj to the plant graph. The ID of the object is set as
        the node label. The object's class is set as the "dexpi_class"
        attribute.

        Parameters
        ----------
        obj : DexpiBaseModel
            DEXPI object to be added to the plant graph.
        """
        attributes = get_data_attributes(obj)
        attributes["dexpi_class"] = obj.__class__.__name__
        self.plant_graph.add_node(obj.id, **attributes)

    def add_edge(
        self,
        source: DexpiBaseModel,
        target: DexpiBaseModel,
        edge: DexpiBaseModel | tuple,
    ):
        """Adds an edge to the plant graph. If the source and target nodes are
        not in the graph yet, create them first. For PipingConnection, add the
        attributes from the PipingNetworkSegmentItem.

        Parameters
        ----------
        source : DexpiBaseModel
            Source node of the edge.
        target : DexpiBaseModel
            Target node of the edge.
        edge : DexpiBaseModel | tuple
            Edge type as a DEXPI class. For PipingConnection, provide a tuple
            with (PipingConnection, PipingNetworkSegmentItem).
        """
        if isinstance(edge, tuple):
            # attributes of pipingNetworkSegment
            attributes = get_data_attributes(edge[1])
            attributes["dexpi_class"] = edge[0].__class__.__name__
        else:
            attributes = get_data_attributes(edge)
            attributes["dexpi_class"] = edge.__class__.__name__

        try:
            self.validate_edge(
                source.__class__.__name__,
                target.__class__.__name__,
                attributes,
                special_edge_attributes=["dexpi_class"],
            )
        except Exception as e:
            warnings.warn(f"Edge cannot be established: {e}")
            return

        self.plant_graph.add_edge(source.id, target.id, **attributes)

    def parse_equipment_and_piping(self):
        """Create networkX graph from equipment and piping"""

        # create dictionary for nozzles of tagged plant items
        nozzles = {}

        # create TaggedPlantItem nodes
        tagged_plant_items = self.plant_model.conceptualModel.taggedPlantItems
        for tagged_plant_item in tagged_plant_items:
            if issubclass(tagged_plant_item.__class__, equipment.NozzleOwner):
                if tagged_plant_item.nozzles:
                    self.add_node(tagged_plant_item)
                    for nozzle in tagged_plant_item.nozzles:
                        nozzles[nozzle] = tagged_plant_item

        # create PipingNetworkSegmentItem nodes
        piping_network_systems = self.plant_model.conceptualModel.pipingNetworkSystems
        for piping_network_system in piping_network_systems:
            piping_network_segments = piping_network_system.segments
            for piping_network_segment in piping_network_segments:
                for item in piping_network_segment.items:
                    self.add_node(item)
                for connection in piping_network_segment.connections:
                    source_item = connection.sourceItem
                    target_item = connection.targetItem
                    if source_item and target_item:
                        if isinstance(source_item, equipment.Nozzle):
                            source_item = nozzles[source_item]
                        if isinstance(target_item, equipment.Nozzle):
                            target_item = nozzles[target_item]
                        self.add_edge(
                            source_item,
                            target_item,
                            (connection, piping_network_segment),
                        )

    def parse_instrumentation(self):
        """Extend graph of equipment and piping with instrumentation."""

        # create and link ControlledActuator nodes
        for actuating_system in self.plant_model.conceptualModel.actuatingSystems:
            valve_reference = actuating_system.operatedValveReference
            valve = valve_reference.valve
            actuator = actuating_system.controlledActuator
            if actuator:
                self.add_node(actuator)
                self.add_edge(actuator, valve, valve_reference)
            positioner = actuating_system.positioner
            if positioner:
                self.add_node(positioner)
                self.add_edge(positioner, valve, valve_reference)

        # create ProcessInstrumentationFunction nodes
        for instrumentation_fn in self.plant_model.conceptualModel.processInstrumentationFunctions:
            self.add_node(instrumentation_fn)
            # create ElectronicFrequencyConverter nodes
            for actuating_e_fn in instrumentation_fn.actuatingElectricalFunctions:
                for actuating_e_s in actuating_e_fn.systems:
                    for e_frequency_converter in actuating_e_s.electronicFrequencyConverter:
                        self.add_node(e_frequency_converter)
                        self.add_edge(
                            instrumentation_fn,
                            e_frequency_converter,
                            instrumentation.ElectronicFrequencyConverter(),
                        )
            # create SignalOffPageConnector nodes
            for signal_opc in instrumentation_fn.signalConnectors:
                self.add_node(signal_opc)
            # create SignalConveyingFunction edges
            for signal_conveying_function in instrumentation_fn.signalConveyingFunctions:
                target = self.determine_reference(signal_conveying_function.target)
                source = self.determine_reference(signal_conveying_function.source)
                self.add_edge(source, target, signal_conveying_function)
            # create nodes for OfflinePrimaryElement
            for process_signal_gen_fn in instrumentation_fn.processSignalGeneratingFunctions:
                if process_signal_gen_fn.systems is not None:
                    for system in process_signal_gen_fn.systems:
                        for primary_element in system.primaryElement:
                            if primary_element.__class__ == instrumentation.OfflinePrimaryElement:
                                self.add_edge(
                                    primary_element,
                                    instrumentation_fn,
                                    instrumentation.SignalLineFunction(),
                                )

    def determine_reference(self, dexpi_object: DexpiBaseModel) -> DexpiBaseModel:
        """Determine the references (Proteus: "association") of an object.
        - SignalConveyingFunction: source and target

        Parameters
        ----------
        dexpi_object : DexpiBaseModel
            DEXPI object that describes reference.

        Returns
        -------
        DexpiBaseModel
            List of DEXPI object to be used as reference.
        """
        if dexpi_object.__class__ == instrumentation.ProcessSignalGeneratingFunction:
            if dexpi_object.sensingLocation.__class__ == equipment.Nozzle:
                for tagged_plant_item in self.plant_model.conceptualModel.taggedPlantItems:
                    for nozzle in tagged_plant_item.nozzles:
                        if nozzle == dexpi_object.sensingLocation:
                            return tagged_plant_item
            elif issubclass(dexpi_object.sensingLocation.__class__, piping.PipingComponent):
                return dexpi_object.sensingLocation
            elif dexpi_object.sensingLocation.__class__ == piping.PipingNetworkSegment:
                raise NotImplementedError(
                    "Select a PipingComponent as the SensingLocation instead of the complete PipingNetworkSystem."
                )
        elif dexpi_object.__class__ == instrumentation.ActuatingFunction:
            return dexpi_object.systems.controlledActuator
        # ProcessInstrumentationFunction, SignalOffPageConnector
        else:
            return dexpi_object

    def draw_process_matplotlib(self) -> Figure:
        """Draw the process graph using matplotlib.pyplot.

        Returns
        -------
        go.Figure
            Figure object for Plotly.
        """
        draw_info = {"node_labels": {}, "node_colors": [], "edge_labels": {}}
        for node in self.plant_graph.nodes:
            class_name = self.plant_graph.nodes[node]["dexpi_class"]
            draw_info["node_labels"][node] = class_name
            if class_name in dir(piping):
                draw_info["node_colors"].append("red")
            elif class_name in dir(instrumentation):
                draw_info["node_colors"].append("blue")
            elif class_name in dir(equipment):
                draw_info["node_colors"].append("green")
        for edge in self.plant_graph.edges:
            draw_info["edge_labels"][edge] = self.plant_graph.edges[edge]["dexpi_class"]
        self.plant_graph = _add_positions(self.plant_graph, len(self.plant_graph.nodes))
        pos = nx.get_node_attributes(self.plant_graph, "pos")
        fig = plt.figure()
        nx.draw(
            self.plant_graph,
            pos,
            labels=draw_info["node_labels"],
            node_color=draw_info["node_colors"],
        )
        nx.draw_networkx_edge_labels(self.plant_graph, pos, edge_labels=draw_info["edge_labels"])
        return fig

    def draw_process_plotly(self) -> go.Figure:
        """Draw the process graph using Plotly.

        Returns
        -------
        go.Figure
            Figure object for Plotly.
        """
        # Initialize the draw info dictionary
        draw_info = {"node_labels": {}, "node_colors": [], "edge_labels": {}}

        # Extract node labels and colors
        for node in self.plant_graph.nodes:
            class_name = self.plant_graph.nodes[node]["dexpi_class"]
            draw_info["node_labels"][node] = class_name
            if class_name in dir(piping):
                draw_info["node_colors"].append("red")
            elif class_name in dir(instrumentation):
                draw_info["node_colors"].append("blue")
            elif class_name in dir(equipment):
                draw_info["node_colors"].append("green")
            else:
                draw_info["node_colors"].append("black")

        # Extract edge labels
        for edge in self.plant_graph.edges:
            draw_info["edge_labels"][edge] = self.plant_graph.edges[edge]["dexpi_class"]

        # Add positions for nodes
        self.plant_graph = _add_positions(
            self.plant_graph, len(self.plant_graph.nodes)
        )  # Or use nx.spring_layout
        pos = nx.get_node_attributes(self.plant_graph, "pos")

        # Create edge traces
        edge_x, edge_y = [], []
        for edge in self.plant_graph.edges:
            x0, y0 = pos[edge[0]]
            x1, y1 = pos[edge[1]]
            edge_x.extend([x0, x1, None])  # None creates a break in the line
            edge_y.extend([y0, y1, None])

        edge_trace = go.Scatter(
            x=edge_x, y=edge_y, line=dict(width=1, color="black"), hoverinfo="none", mode="lines"
        )

        # Create node traces
        node_x, node_y, node_color, node_text = [], [], [], []
        for node in self.plant_graph.nodes:
            x, y = pos[node]
            node_x.append(x)
            node_y.append(y)
            node_color.append(draw_info["node_colors"][list(self.plant_graph.nodes).index(node)])
            node_text.append(draw_info["node_labels"][node])

        node_trace = go.Scatter(
            x=node_x,
            y=node_y,
            mode="markers+text",
            text=node_text,
            textposition="top center",
            marker=dict(size=10, color=node_color, line=dict(width=2)),
        )

        # Create edge label traces
        edge_label_x, edge_label_y, edge_label_text = [], [], []
        for edge, label in draw_info["edge_labels"].items():
            x0, y0 = pos[edge[0]]
            x1, y1 = pos[edge[1]]
            edge_label_x.append((x0 + x1) / 2)  # Midpoint for edge label
            edge_label_y.append((y0 + y1) / 2)
            edge_label_text.append(label)

        edge_label_trace = go.Scatter(
            x=edge_label_x,
            y=edge_label_y,
            mode="text",
            text=edge_label_text,
            textfont=dict(color="black", size=10),
            hoverinfo="none",
        )

        # Combine traces into a Plotly figure
        fig = go.Figure(data=[edge_trace, node_trace, edge_label_trace])

        # Set layout for the figure
        fig.update_layout(
            showlegend=False,
            margin=dict(t=0, l=0, b=0, r=0),
            xaxis=dict(showgrid=False, zeroline=False),
            yaxis=dict(showgrid=False, zeroline=False),
        )

        return fig



from .pydantic_classes import (
    Agglomerator,
    Agitator,
    AgitatorRotor,
    AirCoolingSystem,
    AirEjector,
    AlternatingCurrentGenerator,
    AlternatingCurrentMotor,
    AlternatingCurrentMotorAsComponent,
    AxialBlower,
    AxialCompressor,
    AxialFan,
    BatchWeigher,
    Blower,
    Boiler,
    BriquettingRoller,
    Burner,
    CentrifugalBlower,
    CentrifugalCompressor,
    CentrifugalPump,
    Centrifuge,
    Chamber,
    ChamberOwner,
    Chimney,
    ColumnInternalsArrangement,
    ColumnPackingsArrangement,
    ColumnSection,
    ColumnTraysArrangement,
    CombustionEngine,
    CombustionEngineAsComponent,
    Compressor,
    ContinuousWeigher,
    ConvectionDryer,
    Conveyor,
    CoolingTower,
    CoolingTowerRotor,
    Crusher,
    CrusherElement,
    CustomAgglomerator,
    CustomBlower,
    CustomCentrifuge,
    CustomCompressor,
    CustomCoolingTower,
    CustomDryer,
    CustomElectricGenerator,
    CustomEquipment,
    CustomExtruder,
    CustomFan,
    CustomFilter,
    CustomHeater,
    CustomHeatExchanger,
    CustomMill,
    CustomMixer,
    CustomMobileTransportSystem,
    CustomMotor,
    CustomPump,
    CustomSeparator,
    CustomSieve,
    CustomStationaryTransportSystem,
    CustomTurbine,
    CustomVessel,
    CustomWasteGasEmitter,
    CustomWeigher,
    DirectCurrentGenerator,
    DirectCurrentMotor,
    DirectCurrentMotorAsComponent,
    Displacer,
    DryCoolingTower,
    Dryer,
    DryingChamber,
    EjectorPump,
    ElectricalSeparator,
    ElectricGenerator,
    ElectricHeater,
    Equipment,
    Extruder,
    Fan,
    Feeder,
    Filter,
    FilteringCentrifuge,
    FilteringCentrifugeDrum,
    FilterUnit,
    Flare,
    ForkliftTruck,
    Furnace,
    GasFilter,
    GasTurbine,
    GearBox,
    GravitationalSeparator,
    Grinder,
    GrindingElement,
    HeatedSurfaceDryer,
    Heater,
    HeatExchanger,
    HeatExchangerRotor,
    Impeller,
    Kneader,
    Lift,
    LiquidFilter,
    LoadingUnloadingSystem,
    MechanicalSeparator,
    Mill,
    Mixer,
    MixingElementAssembly,
    MobileTransportSystem,
    Motor,
    MotorAsComponent,
    Nozzle,
    NozzleOwner,
    PackagingSystem,
    PelletizerDisc,
    PlateHeatExchanger,
    PressureVessel,
    ProcessColumn,
    Pump,
    RadialFan,
    RailWaggon,
    ReciprocatingCompressor,
    ReciprocatingExtruder,
    ReciprocatingPressureAgglomerator,
    ReciprocatingPump,
    RevolvingSieve,
    RotaryCompressor,
    RotaryMixer,
    RotaryPump,
    RotatingExtruder,
    RotatingGrowthAgglomerator,
    RotatingPressureAgglomerator,
    Screw,
    ScrubbingSeparator,
    SedimentalCentrifuge,
    SedimentalCentrifugeDrum,
    Separator,
    Ship,
    Sieve,
    SieveElement,
    Silo,
    SpiralHeatExchanger,
    SprayCooler,
    SprayNozzle,
    StaticMixer,
    StationarySieve,
    StationaryTransportSystem,
    SteamGenerator,
    SteamTurbine,
    SubTaggedColumnSection,
    TaggedColumnSection,
    TaggedPlantItem,
    Tank,
    ThinFilmEvaporator,
    TransportableContainer,
    Truck,
    TubeBundle,
    TubularHeatExchanger,
    Turbine,
    Vessel,
    VibratingSieve,
    WasteGasEmitter,
    Weigher,
    WetCoolingTower,
)



"""
Dexpi Piping Toolkit
--------------------
This module contains functions and tools to manipulate the main piping
concepts of a dexpi model. This includes mainly manipulations of piping network
segments.

"""

import warnings
from collections.abc import Callable
from enum import Enum

from pydexpi.dexpi_classes import dexpiModel, piping


class DexpiConnectionException(Exception):
    """Exception for dexpi piping, connection-related errors.

    Raised if a connection or reconnection is attempted that is invalid or may cause corrupt
    segments"""

    pass


class DexpiCorruptPipingSegmentException(Exception):
    """Exception raised if a segment is encountered that violates piping network segment convention."""

    pass


class PipingTraversalException(Exception):
    """Raised if an error occurs during traversal of connected piping components/items/segments."""

    pass


class PipingValidityCode(Enum):
    """
    Enumeration of validity codes for piping related objects.

    Parameters
    ----------
    VALID : int = 0
    INTERNAL_VIOLATION : int = 1
        Violation code for an internal violation in a pipingNetworkSegment.
        Examples are unconnected, loose pipes or items, interrupted segments.
    CONNECTION_CONVENTION_VIOLATION : int = 2
        Violation code if the connection convention regarding a
        pipingNetworkSegment and its pipingConnections is violated
    MULTIPLE_NODE_REFERENCE_VIOLATION : int = 3
        Violation of the convention that a node can only be referenced once
    """

    VALID = 0
    INTERNAL_VIOLATION = 1
    CONNECTION_CONVENTION_VIOLATION = 2
    MULTIPLE_NODE_REFERENCE_VIOLATION = 3


class PipingConnectionConvention(Enum):
    """Convention enumeration for creating connections for higher level piping
    tools where explicit assignment of source and target nodes/items becomes
    impractical.

    Parameters
    ----------
    IN_NODE_0_OUT_NODE_1 : int = 0
        Node with index 0 is used as main inlet, node with index one is main
        outlet
    USE_ITEMS : int = 1
        Use direct assignments to items
    """

    IN_NODE_0_OUT_NODE_1 = 0
    USE_ITEMS = 1


def segment_ends_in_connection(the_segment: piping.PipingNetworkSegment, as_source=False) -> bool:
    """Checks if a segment ends in a connection, as opposed to an item.

    Parameters
    ----------
    the_segment : PipingNetworkSegment
        The segment to be evaluated
    as_source : bool, optional
        Whether the segments source should be checked, by default False

    Returns
    -------
    bool
        True if the segment ends in a connection.
    """
    if not the_segment.connections:
        return False
    destination_item = the_segment.sourceItem if as_source else the_segment.targetItem
    return destination_item not in the_segment.items


def segments_are_connected(
    segment_a: piping.PipingNetworkSegment, segment_b: piping.PipingNetworkSegment
) -> bool:
    """Checks if two segments are connected by comparing their source and target items.

    Parameters
    ----------
    segment_a : PipingNetworkSegment
        The first segment to be checked.
    segment_b : PipingNetworkSegment
        The second segment to be checked.

    Returns
    -------
    bool
        True if the segments are connected, False otherwise.

    Raises
    ------
    ValueError
        If the two segments are the same, as they cannot be connected to themselves.
    """
    # Consistency check if the two segments are the same
    if segment_a == segment_b:
        raise ValueError("Segments are the same. Cannot check if they are connected.")
    # Check if any of the source or target items of one segment are present in the items of the other segment
    return (
        segment_a.sourceItem in segment_b.items
        or segment_a.targetItem in segment_b.items
        or segment_b.sourceItem in segment_a.items
        or segment_b.targetItem in segment_a.items
    )


def connect_piping_network_segment(
    piping_segment: piping.PipingNetworkSegment,
    connector_item: piping.PipingSourceItem | piping.PipingTargetItem,
    connector_node_index: int | None = None,
    as_source=False,
    force_reconnect=False,
) -> None:
    """
    Connects a piping network segment by inferring the pipe to reconnect.

    This is based on the assumption that a valid piping network segment has at
    most one unconnected incoming and at most one unconnected outflowing
    connection. Note that a valid piping network segment may end on a segment
    item like a valve, in which case there are no outgoing flow connections
    (same for input). In this case, a ValueError is raised. An Exception is also
    raised if the piping network segment is corrupted (i.e., has more than one
    inflowing or outflowing connection).

    Parameters
    ----------
    piping_segment : PipingNetworkSegment
        Piping segment to be reconnected.
    connector_item : Union[PipingSourceItem, PipingTargetItem]
        The connector object to which the segment will be connected.
        It can be either a PipingNode, PipingSourceItem, or PipingTargetItem.
    connector_node_index: int, optional
        Optionally, the index of the node of connector_item to be connected to,
        by default None
    as_source : bool, optional
        If True, the connection is established as source, else as target.
        Defaults to False.
    force_reconnect : bool, optional
        If False, raises an exception if the piping connection is already
        connected. Defaults to False.

    Returns
    -------
    None
        This function modifies the state of the piping segment and the
        connection but does not return any value.

    Raises
    ------
    DexpiConnectionException
        If the piping connection is already connected, and force_reconnect is
        False.
    ValueError
        If the piping network segment either has no unconnected output
        connection (or input connection if as_source) or has more than one such
        connections, in which case it is corrupted.
    DexpiCorruptPipingSegmentException
        If the segment to be connected is found to be corrupt
    """
    # Find last connection and item. Last item is None if its None alltogether
    # OR if it isnt among the segment items internally
    if piping_segment.connections:
        last_connection = find_final_connection(piping_segment, as_source=as_source)
        last_item = last_connection.sourceItem if as_source else last_connection.targetItem
        last_item = last_item if last_item in piping_segment.items else None

    # Manage case segment has no connection (is empty or consists of one item):
    else:
        last_connection = None
        if not piping_segment.items:
            last_item = None
        elif len(piping_segment.items) == 1:
            last_item = piping_segment.items[0]
        else:
            msg = (
                f"The segment {piping_segment} has multiple items but no "
                f"connections and may be corrupt"
            )
            raise DexpiCorruptPipingSegmentException(msg)

    # Determine if connection references an internal object, in which case a
    # a reconnection would be invalid and cause a corrupt segment
    if last_item is not None:
        msg = (
            f"Piping segment {piping_segment} references internal "
            f"object as {'source' if as_source else 'target'}. Reconnect "
            f"will cause corrupting the segment."
        )
        raise ValueError(msg)

    if connector_node_index is not None:
        connector_node = connector_item.nodes[connector_node_index]
    else:
        connector_node = None

    _connect_piping_connection(
        connector_item,
        last_connection,
        piping_segment,
        connector_node,
        as_source,
        force_reconnect,
    )


def construct_new_segment(
    segment_items: list[piping.PipingNetworkSegmentItem],
    segment_connections: list[piping.PipingConnection],
    source_connector_item: piping.PipingSourceItem | piping.PipingTargetItem = None,
    target_connector_item: piping.PipingSourceItem | piping.PipingTargetItem = None,
    source_connector_node_index: int = None,
    target_connector_node_index: int = None,
    connectivity_convention: PipingConnectionConvention = PipingConnectionConvention.IN_NODE_0_OUT_NODE_1,
    **segment_kwargs,
) -> piping.PipingNetworkSegment:
    """Constructs a new segment from the segment_items and segment_connections
    passed.

    Segment items and connections need to be in the order they occurr in the
    segment. The shape at the beginning and end of the segment is inferred by
    the source and target object specified. If no Source is given or the source
    is external, the segment starts with a pipe. Same pattern for target.
    However, in this case, the lengths of the passed objects need to match
    accordingly. Warnings printed in case the type of the source and target dont
    match the specified convention

    Parameters
    ----------
    segment_items : list[PipingNetworkSegmentItem]
        The items of the segments that are to be included in order of
        occurrence.
    segment_connections : list[PipingConnection]
        The connections of the segments that are to be included in order of
        occurrence.
    connectivity_convention : PipingConnectionConvention
        The convention to be applied when stringing together the connections and
        items.
    source_connector_item : Union[PipingSourceItem, PipingTargetItem], optional
        The item to be specified as source, can be first element of the
        segment, in which case no first connection is assigned, by default None.
    target_connector_item : Union[PipingSourceItem, PipingTargetItem], optional
        The item to be specified as target, can be last element of the
        segment, in which case no final connection is assigned, by default None.
    source_connector_node_index : int, optional
        The index to the node to be specified as source in the source item,
        by default None.
    target_connector_node_index : int, optional
        The index to the node to be specified as target in the target item,
        by default None.
    **segment_kwargs:
        The additional quargs to be passed on to the constructor of the segment.

    Returns
    -------
    PipingNetworkSegment
        The new piping network segment created

    Raises
    ------
    ValueError
        Raised if the arguments are invalid, such as if an internal item is
        referenced as source or target or if the lengthes of the items and
        connections dont line up with regards to the source and target passed.
    """
    # First, screen for invalid sources
    invalid_sources = segment_items[1:]
    if source_connector_item in invalid_sources:
        msg = f"Cant assign source item to member item {source_connector_item} except the first one"
        raise ValueError(msg)
    invalid_targets = segment_items[:-1]
    if target_connector_item in invalid_targets:
        msg = f"Cant assign target item to member item {target_connector_item} except the final one"
        raise ValueError(msg)

    if connectivity_convention == PipingConnectionConvention.IN_NODE_0_OUT_NODE_1:
        if source_connector_node_index != 0 and source_connector_item is not None:
            warnings.warn(
                "With convention IN_NODE_0_OUT_NODE_1, the correct source node index should be 0"
            )
        if target_connector_node_index != 1 and target_connector_item is not None:
            warnings.warn(
                "With convention IN_NODE_0_OUT_NODE_1, the correct target node index should be 1"
            )

    # Assign nodes
    if source_connector_item is not None and source_connector_node_index is not None:
        source_connector_node = source_connector_item.nodes[source_connector_node_index]
    elif source_connector_node_index is not None:
        msg = "Cannot pass a source outside connector node index without a source connector item"
        raise ValueError(msg)
    else:
        source_connector_node = None

    if target_connector_item is not None and target_connector_node_index is not None:
        target_connector_node = target_connector_item.nodes[target_connector_node_index]
    elif target_connector_node_index is not None:
        msg = "Cannot pass a target outside connector node index without a target connector item"
        raise ValueError(msg)
    else:
        target_connector_node = None

    # Examine source and target connector object and append to segment kwargs.
    # Warnings issued in case the type of the source and target dont match the
    # specified convention
    for type_string, outside_connector_item, outside_connector_node in zip(
        ["source", "target"],
        [source_connector_item, target_connector_item],
        [source_connector_node, target_connector_node],
    ):
        warn = False

        if outside_connector_item is not None:
            if (
                connectivity_convention == PipingConnectionConvention.IN_NODE_0_OUT_NODE_1
                and outside_connector_node is None
            ):
                warn = True
            kwargs_key = type_string + "Item"
            segment_kwargs[kwargs_key] = outside_connector_item
        if outside_connector_node is not None:
            if connectivity_convention == PipingConnectionConvention.USE_ITEMS:
                warn = True
            kwargs_key = type_string + "Node"
            segment_kwargs[kwargs_key] = outside_connector_node

        if warn:
            warnings.warn(
                f"The {type_string} connection specified via the arguments "
                f"violates the indicated convention "
                f"{connectivity_convention}"
            )

    # Check if source and target are internals
    source_is_internal = source_connector_item in segment_items
    target_is_internal = target_connector_item in segment_items
    if segment_items:
        if source_is_internal and source_connector_item != segment_items[0]:
            msg = f"Specified source {source_connector_item} in the segment, but not the first element of segment_items"
            raise ValueError(msg)
        if target_is_internal and target_connector_item != segment_items[-1]:
            msg = f"Specified source {source_connector_item} in the segment, but not the first element of segment_items"
            raise ValueError(msg)

    if source_is_internal and target_is_internal:
        if len(segment_items) != len(segment_connections) + 1:
            msg = (
                "With both source and target internal, the number of "
                "items needs to be one more than the connections"
            )
            raise ValueError(msg)
    elif source_is_internal ^ target_is_internal:
        if len(segment_items) != len(segment_connections):
            msg = (
                "With either the source or target internal, the "
                "number of connections needs to equal the number of items"
            )
            raise ValueError(msg)
    elif (not source_is_internal) and (not target_is_internal):
        if len(segment_items) + 1 != len(segment_connections):
            msg = (
                "With both source and target not internal, the number "
                "of items needs to be one less than the connections"
            )
            raise ValueError(msg)

    # Connect each connection of the segment
    for index, connection in enumerate(segment_connections):
        item_index = index - 1 if source_is_internal else index
        # Connect source of connection
        if index == 0 and not source_is_internal:
            _connect_piping_connection(
                source_connector_item,
                piping_connection=connection,
                connector_node=source_connector_node,
                as_source=True,
            )
        else:
            source_item = segment_items[item_index - 1]
            if connectivity_convention == PipingConnectionConvention.IN_NODE_0_OUT_NODE_1:
                source_node = source_item.nodes[1]
            elif connectivity_convention == PipingConnectionConvention.USE_ITEMS:
                source_node = None
            _connect_piping_connection(
                source_item,
                piping_connection=connection,
                connector_node=source_node,
                as_source=True,
            )

        # Connect target of connection
        if index == len(segment_connections) - 1 and not target_is_internal:
            _connect_piping_connection(
                target_connector_item,
                piping_connection=connection,
                connector_node=target_connector_node,
            )
        else:
            target_item = segment_items[item_index]
            if connectivity_convention == PipingConnectionConvention.IN_NODE_0_OUT_NODE_1:
                target_node = target_item.nodes[0]
            elif connectivity_convention == PipingConnectionConvention.USE_ITEMS:
                target_node = None
            _connect_piping_connection(
                target_item,
                piping_connection=connection,
                connector_node=target_node,
            )

    # Create the piping network segment and returns it
    new_segment = piping.PipingNetworkSegment(
        items=segment_items, connections=segment_connections, **segment_kwargs
    )
    return new_segment


def construct_new_segment_already_connected(
    segment_items: list[piping.PipingNetworkSegmentItem],
    segment_connections: list[piping.PipingConnection],
    source_node_index: int | None = None,
    target_node_index: int | None = None,
    **segment_kwargs,
) -> piping.PipingNetworkSegment:
    """Constructs a new piping network segment from already connected items and connections.

    Items and connections are first ordered according to their connectivity. If the connections are
    invalid, ValueErrors are raised by the sort_connected_items_and_connections function.

    Parameters
    ----------
    segment_items : list[piping.PipingNetworkSegmentItem]
        The list of items that are part of the segment.
    segment_connections : list[piping.PipingConnection]
        The list of connections that are part of the segment.
    source_node_index : int|None, optional
        The index of the source node in the source item. Only used if the target item is internal,
        by default None.
    target_node_index : int|None, optional
        The index of the target node in the target item. Only used if the target item is internal,
        by default None.

    Returns
    -------
    piping.PipingNetworkSegment
        The newly constructed piping network segment.

    Raises
    ------
    ValueError
        If the items and connections are not connected in a valid way.
    """

    # First, order the segment connections and items
    sorted_items, sorted_connections = sort_connected_items_and_connections(
        segment_items, segment_connections
    )

    # Create the piping network segment
    if sorted_connections:
        source_item = sorted_connections[0].sourceItem
        target_item = sorted_connections[-1].targetItem
    else:
        # In this case, there should only be one item in the segment
        if len(sorted_items) != 1:
            raise ValueError("No connections and more than one item in the segment.")
        source_item = sorted_items[0]
        target_item = sorted_items[0]
    new_segment = piping.PipingNetworkSegment(
        items=sorted_items,
        connections=sorted_connections,
        sourceItem=source_item,
        targetItem=target_item,
        **segment_kwargs,
    )

    # Assign segment source and target nodes: These are the same as the source and target nodes of
    # the first and last connection, unless these are nodes in internal items. In that case, assign
    # the nodes according to the piping convention:
    if source_item in new_segment.items:
        if source_node_index is not None:
            new_segment.sourceNode = source_item.nodes[source_node_index]
        else:
            pass  # No assignment needs to be done, as the source node is already None
    else:
        if source_node_index is not None:
            raise ValueError(
                "Source node index is not None, but node already inferred from connection."
            )
        new_segment.sourceNode = sorted_connections[0].sourceNode

    if target_item in new_segment.items:
        if target_node_index is not None:
            new_segment.targetNode = target_item.nodes[1]
        else:
            pass  # No assignment needs to be done, as the target node is already None
    else:
        if target_node_index is not None:
            raise ValueError(
                "Target node index is not None, but node already inferred from connection."
            )
        new_segment.targetNode = sorted_connections[-1].targetNode

    return new_segment


def insert_item_to_segment(
    the_segment: piping.PipingNetworkSegment,
    position: int | piping.PipingNetworkSegmentItem,
    the_item: piping.PipingNetworkSegmentItem,
    the_connection: piping.PipingConnection,
    item_target_node_index: int = None,
    item_source_node_index: int = None,
    insert_before: bool = False,
) -> None:
    """Inserts a piping item and a piping connection into a piping segment
    before or after a member connection. Intended for segments that are fully
    connected.

    Parameters
    ----------
    the_segment : pipingNetworkSegment
        The segment into which an item and a connection are to be inserted.
    position : Union[int, PipingNetworkSegmentItem]
        Position before or after whicht the insertion happens. Can either be the
        existing item object or the integer index of the item.
    the_connection : PipingConnection
        The connection to be inserted.
    the_item : PipingNetworkSegmentItem
        The item to be inserted
    item_target_node_index : int, optional
        The target node index of the_item to be used for target connection if desired.
        By default None.
    item_source_node_index : int, optional
        The source node index of the_item to be used for source connection if desired.
        By default None.
    insert_before : bool, optional
        If the objects are inserted before the item object, else after.
        By default False.

    Returns
    -------
    None: the_segment is manipulated in place.

    Raises
    ------
    ValueError
        If the arguments passed are not appropriately associated, e.g. if the
        source and target nodes are out of bounds or the position
        object isnt associated to the segment.
    """
    # Some consistency checks on passed arguments
    # Validity check if item not already in the segment
    if the_item in the_segment.items:
        msg = f"Item {the_item} is already member of {the_segment}."
        raise ValueError(msg)
    # Validity check if connection not already in the segment
    if the_connection in the_segment.items:
        msg = f"Connection {the_connection} is already member of {the_segment}."
        raise ValueError(msg)

    for node_index in [item_source_node_index, item_target_node_index]:
        if node_index is not None and node_index >= len(the_item.nodes):
            msg = f"Index {node_index} out of bounds for nodes of item {the_item}"
            raise ValueError(msg)
    # Validity check if the position is in fact a member of the segment if it
    # is a segment item
    if isinstance(position, piping.PipingNetworkSegmentItem) and position not in the_segment.items:
        msg = f"Piping item {position} not associated to segment  {the_segment}."
        raise ValueError(msg)

    # Manage assignemnt of item index and the actual item depending on the type
    # specified in the position argument
    if isinstance(position, piping.PipingNetworkSegmentItem):
        item_at_position = position
        item_index = the_segment.items.index(position)
    else:
        if the_segment.items:
            item_at_position = the_segment.items[position]
        elif position == 0:
            item_at_position = None
        else:
            msg = f"Segment {the_segment} has no items, so position index {position} is invalid."
        item_index = position

    # Find connection at position
    connection_at_position = None
    insertion_at_segment_end = False
    # Manage case of a single pipe segment:
    if not the_segment.items and the_segment.connections:
        connection_at_position = the_segment.connections[0]
    # Otherwise, find connection that currently connects to item at position for later reconnection
    elif the_segment.items:
        for connection in the_segment.connections:
            connection_enditem = connection.targetItem if insert_before else connection.sourceItem
            if connection_enditem == item_at_position:
                connection_at_position = connection
                break
    # Case of a segment without items or connections
    else:
        connection_at_position = None

    # If no connection at position is found, this means we are inserting at segment end or the segment is corrupt
    if connection_at_position is None:
        if (insert_before and item_index == 0) or (
            not insert_before and item_index == len(the_segment.items)
        ):
            insertion_at_segment_end = True
        else:
            msg = (
                f"No connection found connecting to the item {item_at_position} at given position "
                f"and item is not at the end. Segment {the_segment} may be corrupt."
            )
            raise DexpiCorruptPipingSegmentException(msg)

    if insert_before:
        # Connect the_item with the_connection
        the_connection.sourceItem = the_item
        if item_source_node_index is not None:
            the_connection.sourceNode = the_item.nodes[item_source_node_index]
        else:
            the_connection.sourceNode = None

        if not insertion_at_segment_end:
            # Connect piping target to the original target of the obj_at_position
            the_connection.targetNode = connection_at_position.targetNode
            the_connection.targetItem = connection_at_position.targetItem
            # Reconnect the connection_at_position target
            connection_at_position.targetItem = the_item
            if item_target_node_index is not None:
                connection_at_position.targetNode = the_item.nodes[item_target_node_index]
            else:
                connection_at_position.targetNode = None
            # Define connection index for insertion of new objects
            connection_index = the_segment.connections.index(connection_at_position) + 1
        else:
            # Connect piping target to the previous source of the segment
            the_connection.targetNode = the_segment.sourceNode
            the_connection.targetItem = the_segment.sourceItem
            # Reconnect the segment source (=target of the item as per Dexpi convention)
            the_segment.sourceItem = the_item
            if item_target_node_index is not None:
                the_segment.sourceNode = the_item.nodes[item_target_node_index]
            else:
                the_segment.sourceNode = None
            connection_index = 0
    else:
        # Connect the_item with the_connection
        the_connection.targetItem = the_item
        if item_target_node_index is not None:
            the_connection.targetNode = the_item.nodes[item_target_node_index]
        else:
            the_connection.targetNode = None
        if not insertion_at_segment_end:
            # Connect piping source to the original source of the obj_at_position
            the_connection.sourceNode = connection_at_position.sourceNode
            the_connection.sourceItem = connection_at_position.sourceItem
            # Reconnect the piping_obj source
            connection_at_position.sourceItem = the_item
            if item_source_node_index is not None:
                connection_at_position.sourceNode = the_item.nodes[item_source_node_index]
            else:
                connection_at_position.sourceNode = None

            # Define connection index for insertion of new objects
            connection_index = the_segment.connections.index(connection_at_position)
            item_index += 1
        else:
            # Connect piping source to the previous target of the segment
            the_connection.sourceNode = the_segment.targetNode
            the_connection.sourceItem = the_segment.targetItem
            # Reconnect the segment target (=source of the item as per Dexpi convention)
            the_segment.targetItem = the_item
            if item_source_node_index is not None:
                the_segment.targetNode = the_item.nodes[item_source_node_index]
            else:
                the_segment.targetNode = None
            connection_index = len(the_segment.connections)

    # Insert the objects
    the_segment.connections.insert(connection_index, the_connection)
    the_segment.items.insert(item_index, the_item)


def append_item_to_unconnected_segment(
    the_segment: piping.PipingNetworkSegment,
    the_item: piping.PipingNetworkSegmentItem,
    node_index_for_connection: int = None,
    node_index_segment_end: int = None,
    insert_before: bool = False,
) -> None:
    """Appends a piping network segment item to a free piping network segment.

    Parameters
    ----------
    the_segment : PipingNetworkSegment
        The segment to be appended to.
    the_item : PipingNetworkSegmentItem
        The item to be appended.
    node_index_for_connection : int, optional
        The node of the item that should be used for connection, if desired.
        If segment is empty, this node becomes the opposite segment end to
        node_index_segment_end, by default None.
    node_index_segment_end : int, optional
        The node of the item that is used as the new segment end.
        If none, item is used as reference directly, by default None.
    insert_before : bool, optional
        If the item should be appended at the beginning of the segment,
        otherwise the end, by default False. (If segment is empty,
        insert_before=True switches the role of the node indexes)

    Returns
    -------
    None: the_segment is manipulated in place.

    Raises
    ------
    ValueError:
        If connection already in the segment

    ValueError:
        If segment isnt unconnected or not suitable to append an item in that
        position, or if the node is not associated th the object.

    ValueError:
        If a second_node_single_item is passed for segments with items already
        in them

    """
    # Validity check if item not already in the segment
    if the_item in the_segment.items:
        msg = f"Item {the_item} is already member of {the_segment}."
        raise ValueError(msg)

    # Validity check if the segment is unconnected
    if insert_before:
        if the_segment.sourceItem is not None:
            msg = (
                f"Segment {the_segment} has a source and isn't unconnected. "
                f"Consider using insert_item_to_segment instead"
            )
            raise ValueError(msg)
    else:
        if the_segment.targetItem is not None:
            msg = (
                f"Segment {the_segment} has a target and isn't unconnected. "
                f"Consider using insert_item_to_segment instead"
            )
            raise ValueError(msg)

    # Find last connection and item. Last item is found as the segment endpoint
    # with segment items as candidates, which returns None if its None
    # alltogether OR if it isnt among the segment items internally
    # (the latter requires the segment connection check above)
    if the_segment.connections:
        last_connection = find_final_connection(the_segment, as_source=insert_before)
        last_item = last_connection.sourceItem if insert_before else last_connection.targetItem
        last_item = last_item if last_item in the_segment.items else None

    # Manage case segment has no connection (is empty or consists of one item):
    else:
        last_connection = None
        if not the_segment.items:
            last_item = None
        elif len(the_segment.items) == 1:
            last_item = the_segment.items[0]
        else:
            msg = (
                f"The segment {the_segment} has multiple items but no "
                f"connections and may be corrupt"
            )
            raise DexpiCorruptPipingSegmentException(msg)

    # Validity check if final object is a connection and can be appended
    if last_item is not None:
        raise ValueError(
            f"Final object in segment {the_segment} is an item. Cannot add new item {the_item}"
        )

    # Determine connectors
    internal_connector_item = the_item
    if node_index_for_connection is not None:
        internal_connector_node = the_item.nodes[node_index_for_connection]
    else:
        internal_connector_node = None
    segment_connector_item = the_item
    if node_index_segment_end is not None:
        segment_connector_node = the_item.nodes[node_index_segment_end]
    else:
        segment_connector_node = None

    if last_connection is not None:
        _connect_piping_connection(
            internal_connector_item,
            piping_connection=last_connection,
            connector_node=internal_connector_node,
            as_source=insert_before,
        )
    # Case segment has no connection and needs the other end reconnected to the item too.
    else:
        _connect_piping_connection(
            segment_connector_item,
            piping_segment=the_segment,
            connector_node=segment_connector_node,
            as_source=not insert_before,
        )

    _connect_piping_connection(
        segment_connector_item,
        piping_segment=the_segment,
        connector_node=segment_connector_node,
        as_source=insert_before,
    )

    if insert_before:
        the_segment.items.insert(0, the_item)
    else:
        the_segment.items.append(the_item)


def append_connection_to_unconnected_segment(
    the_segment: piping.PipingNetworkSegment,
    the_connection: piping.PipingConnection,
    node_index_for_connection: int = None,
    insert_before: bool = False,
) -> None:
    """Appends a connection object to a free piping network segment.

    Parameters
    ----------
    the_segment : PipingNetworkSegment
        The segment to be appended to.
    the_connection : PipingConnection
        The connection to be appended.
    node_index_for_connection : int, optional
        The node index of the final item in the segment that should be used for
        connection if desired, by default None.
    insert_before : bool, optional
        If the connection should be appended at the beginning of the segment,
        otherwise the end, by default False.

    Returns
    -------
    None: the_segment is manipulated in place.

    Raises
    ------
    ValueError:
        If connection already in the segment

    ValueError:
        If segment isnt unconnected or not suitable to append a connection in
        that position, or if the node is not associated th the object.

    """
    # Validity check if connection not already in the segment
    if the_connection in the_segment.connections:
        msg = f"Connection {the_connection} is already member of {the_segment}."
        raise ValueError(msg)
    # Find last connection and item. Last item is found as the segment endpoint
    # with segment items as candidates, which returns None if its None
    # alltogether OR if it isnt among the segment items internally
    if the_segment.connections:
        last_connection = find_final_connection(the_segment, as_source=insert_before)
        last_item = last_connection.sourceItem if insert_before else last_connection.targetItem
        last_item = last_item if last_item in the_segment.items else None

    # Manage case segment has no connection (is empty or consists of one item):
    else:
        last_connection = None
        if not the_segment.items:
            last_item = None
        elif len(the_segment.items) == 1:
            last_item = the_segment.items[0]
        else:
            msg = (
                f"The segment {the_segment} has multiple items but no "
                f"connections and may be corrupt"
            )
            raise DexpiCorruptPipingSegmentException(msg)

    # Validity check if the segment is unconnected and ends in an item
    if last_item is None and the_segment.items:
        msg = (
            f"Final object in segment {the_segment} is a connection or "
            f"outside of the segment (meaning it's already connected)."
            f"In both cases: Can't add connection {the_connection}"
        )
        raise ValueError(msg)
    connector_item = last_item
    if node_index_for_connection is not None and last_item is not None:
        connector_node = last_item.nodes[node_index_for_connection]
    else:
        connector_node = None

    # Connect
    if insert_before:
        _connect_piping_connection(
            connector_item,
            the_connection,
            connector_node=connector_node,
            as_source=False,
        )
        the_segment.connections.insert(0, the_connection)
        the_segment.sourceNode = the_connection.sourceNode
        the_segment.sourceItem = the_connection.sourceItem
    else:
        _connect_piping_connection(
            connector_item,
            the_connection,
            connector_node=connector_node,
            as_source=True,
        )
        the_segment.connections.append(the_connection)
        the_segment.targetNode = the_connection.targetNode
        the_segment.targetItem = the_connection.targetItem


def find_final_connection(
    the_segment: piping.PipingNetworkSegment, as_source: bool = False
) -> piping.PipingConnection:
    """Finds the first or last connection object in the_segment. It does so by
    comparing the source and target objects of the segment and the connections.
    Note that in a valid segment, these should be the first or final connection.
    However, this method includes some validity checks, so it is recommended to
    use this instead.

    Parameters
    ----------
    the_segment : pipingNetworkSegment
        The segment whose final connection is to befound.
    as_source : bool, optional
        If the first connection should be found, else last, by default False.

    Returns
    -------
    pipingConnection
        The first or last connection.

    Raises
    ------
    DexpiCorruptPipingSegmentException
        If no connection with the same source and target info as the segment is
        found.
    """
    if not the_segment.connections:
        raise ValueError(f"Segment {the_segment} doesn't have any connections")
    final_connection = None
    # Find the connection with the same source/target info as segment
    for connection in the_segment.connections:
        if as_source:
            if connection.sourceItem is the_segment.sourceItem:
                final_connection = connection
        else:
            if connection.targetItem is the_segment.targetItem:
                final_connection = connection
    # Raise exception if None was found
    if final_connection is None:
        msg = (
            f"Piping segment {the_segment} has no connections with the same "
            f"{'source' if as_source else 'target'} and may be corrupt."
        )
        raise DexpiCorruptPipingSegmentException(msg)

    final_connection_index = the_segment.connections.index(final_connection)
    what_index_should_be = 0 if as_source else len(the_segment.connections) - 1
    if final_connection_index != what_index_should_be:
        msg = (
            f"The final connection found in segment {the_segment} is not in the"
            f"correct position {'0' if as_source else '-1'}."
        )
        raise DexpiCorruptPipingSegmentException(msg)

    return final_connection


def get_unconnected_piping_segments(
    conceptual_model: dexpiModel.ConceptualModel, as_source=False
) -> list[piping.PipingNetworkSegment]:
    """
    Returns a list of unconnected piping segments ending in a connection in the conceptual model.

    Parameters
    ----------
    conceptual_model : ConceptualModel
        The model to be analyzed.
    as_source : bool, optional
        Whether to look for unconnected sources or targets in piping segments.
        Defaults to False.

    Returns
    -------
    list[PipingNetworkSegment]
        The list of identified unconnected segments.
    """
    unconnected_segments = []
    for piping_network_system in conceptual_model.pipingNetworkSystems:
        for piping_network_segment in piping_network_system.segments:
            if as_source:
                if piping_network_segment.sourceItem is None:
                    unconnected_segments.append(piping_network_segment)
            else:
                if piping_network_segment.targetItem is None:
                    unconnected_segments.append(piping_network_segment)

    return unconnected_segments


def traverse_items_and_connections(
    all_items: list[piping.PipingNetworkSegmentItem],
    all_connections: list[piping.PipingConnection],
    start_element: piping.PipingNetworkSegmentItem | piping.PipingConnection,
    end_condition: piping.PipingNetworkSegmentItem
    | piping.PipingConnection
    | Callable[[piping.PipingNetworkSegmentItem | piping.PipingConnection], bool],
    reverse=False,
) -> list[piping.PipingNetworkSegmentItem | piping.PipingConnection]:
    """
    Traverses piping items and connections until the end condition is met.

    The traversal is done in a depth-first manner, starting from the specified start element.
    Branching decisions are inferred from the order of the connections in the list of all connections.
    Branches that start with connections that are higher in the list are traversed first.
    The traversal can be done in reverse order by setting the reverse parameter to True.

    Parameters
    ----------
    all_items : list[piping.PipingNetworkSegmentItem]
        List of all items in the piping network segment.
    all_connections : list[piping.PipingConnection]
        List of all connections in the piping network segment.
    start_element : piping.PipingNetworkSegmentItem|piping.PipingConnection
        The element from which to start the traversal.
    end_condition : piping.PipingNetworkSegmentItem|piping.PipingConnection|
                    Callable[[piping.PipingNetworkSegmentItem|piping.PipingConnection], bool]
        The condition that determines when to stop the traversal. This can be a specific element to
        be encountered or a callable function for more complex end conditions.
    reverse : bool, optional
        Whether to traverse in reverse order. Defaults to False.

    Returns
    -------
    list[piping.PipingNetworkSegmentItem|piping.PipingConnection]
        A list of traversed elements.

    Raises
    ------
    PipingTraversalException
        If a circle is encountered or if the traversal cannot continue due to invalid connections.
    """

    # Initialize the traversal
    traversed_elements: list[piping.PipingNetworkSegmentItem | piping.PipingConnection] = [
        start_element
    ]

    # Determine if the start element is a connection or an item
    if isinstance(start_element, piping.PipingNetworkSegmentItem):
        last_was_connection = False
        current_item = start_element
        current_connection = None
    elif isinstance(start_element, piping.PipingConnection):
        last_was_connection = True
        current_item = None
        current_connection = start_element
    else:
        raise TypeError(
            "start_element must be either a PipingNetworkSegmentItem or a PipingConnection"
        )

    # Check if the end condition is a callable function or a specific item/connection.
    # If it's a specific item/connection, create a function to check for equality.
    if isinstance(end_condition, piping.PipingNetworkSegmentItem | piping.PipingConnection):
        end_element = end_condition

        def end_condition_func(
            element: piping.PipingNetworkSegmentItem | piping.PipingConnection,
        ) -> bool:
            return element == end_element

        end_condition = end_condition_func

    if end_condition(start_element):
        return traversed_elements

    # Begin the traversal loop
    while True:
        # Case: Next is an item (if last_was_connection is True)
        if last_was_connection:
            # Get next item
            next_item = (
                current_connection.targetItem if not reverse else current_connection.sourceItem
            )

            # Check if the next item is invalid for further traversal
            if next_item is None:
                msg = f"The connection has no {'source' if reverse else 'target'} item. Cannot traverse further."
                raise PipingTraversalException(msg)
            if next_item not in all_items:
                msg = "The target item is not in the list of all items. Cannot traverse further."
                raise PipingTraversalException(msg)
            if next_item in traversed_elements:
                msg = "Circle encountered: The target item has already been traversed. Cannot traverse further."
                raise PipingTraversalException(msg)

            # Append the next item to the traversed elements
            traversed_elements.append(next_item)
            current_item = next_item
            last_was_connection = False

            # Check if the end condition is met. If so, break the loop.
            if end_condition(current_item):
                break

        else:
            # Find all connections originating from the current item
            next_connections = []
            for connection in all_connections:
                connection_item = connection.sourceItem if not reverse else connection.targetItem
                if connection_item == current_item:
                    next_connections.append(connection)

            # If no connections are found, traversal cannot continue
            if not next_connections:
                msg = "The item has no source connections. Cannot traverse further."
                raise PipingTraversalException(msg)

            # If multiple connections are found, handle branching
            elif len(next_connections) > 1:
                branch_traversal = None

                # Attempt to traverse each branch until a valid path is found
                for connection in next_connections:
                    try:
                        # Filter out already traversed items and connections
                        remaining_items = [
                            item for item in all_items if item not in traversed_elements
                        ]
                        remaining_connections = [
                            con for con in all_connections if con not in traversed_elements
                        ]

                        # Recursively traverse the branch
                        branch_traversal = traverse_items_and_connections(
                            remaining_items, remaining_connections, connection, end_condition
                        )
                        # Stop if a valid branch is found
                        break

                    except PipingTraversalException:
                        continue

                # If a valid branch is found, append it to the traversed elements. Otherwise, raise
                # an exception.
                if branch_traversal is not None:
                    return traversed_elements + branch_traversal
                else:
                    msg = "No valid branch found. Cannot traverse further."
                    raise PipingTraversalException(msg)

            # If only one connection is found, continue traversal
            else:
                next_connection = next_connections[0]

                # Check for circular references
                if next_connection in traversed_elements:
                    msg = "Circle encountered: The connection has already been traversed. Cannot traverse further."
                    raise PipingTraversalException(msg)

            # Append the next connection to the traversed elements
            traversed_elements.append(next_connection)
            current_connection = next_connection
            last_was_connection = True

            # Check if the end condition is met. If so, break the loop.
            if end_condition(current_connection):
                break

    # Return the list of traversed elements
    return traversed_elements


def sort_connected_items_and_connections(
    items: list[piping.PipingNetworkSegmentItem], connections: list[piping.PipingConnection]
) -> tuple[list[piping.PipingNetworkSegmentItem], list[piping.PipingConnection]]:
    """
    Sorts the items and connections of a piping network segment in the order of occurrence.

    Parameters
    ----------
    items : list[piping.PipingNetworkSegmentItem]
        The list of items to be sorted.
    connections : list[piping.PipingConnection]
        The list of connections to be sorted.

    Returns
    -------
    tuple[list[piping.PipingNetworkSegmentItem], list[piping.PipingConnection]]
        A tuple containing the sorted items and connections.

    Raises
    ------
    ValueError
        If the segment is not validly connected, such as having multiple starting points
        or disconnected elements.
    """
    # Find the starting element (either an item or a connection)
    start_element = None
    for connection in connections:
        if connection.sourceItem not in items:
            if start_element is None:
                start_element = connection
            else:
                raise ValueError(
                    "Multiple connections found with no source item. Segment is not validly connected."
                )
    if start_element is None:
        for item in items:
            if item not in [connection.targetItem for connection in connections]:
                if start_element is None:
                    start_element = item
                else:
                    raise ValueError(
                        "Multiple items found that are not referenced by any connection. "
                        "Segment is not validly connected."
                    )

    if start_element is None:
        raise ValueError("No valid starting element found. Segment is not validly connected.")

    # Find end element (either an item or a connection)
    end_element = None
    for connection in connections:
        if connection.targetItem not in items:
            if end_element is None:
                end_element = connection
            else:
                raise ValueError(
                    "Multiple connections found with no target item. Segment is not validly connected."
                )
    if end_element is None:
        for item in items:
            if item not in [connection.sourceItem for connection in connections]:
                if end_element is None:
                    end_element = item
                else:
                    raise ValueError(
                        "Multiple items found that are not referenced as a source by any connection. "
                        "Segment is not validly connected."
                    )

    if end_element is None:
        raise ValueError("No valid ending element found. Segment is not validly connected.")

    # Traverse the segment to determine the order of items and connections
    try:
        traversed_elements = traverse_items_and_connections(
            items, connections, start_element, end_element
        )
    except PipingTraversalException as e:
        raise ValueError(f"Traversal failed: {e}")

    # Separate the traversed elements into items and connections
    ordered_items = [
        element
        for element in traversed_elements
        if isinstance(element, piping.PipingNetworkSegmentItem)
    ]
    ordered_connections = [
        element for element in traversed_elements if isinstance(element, piping.PipingConnection)
    ]

    if len(ordered_items) != len(items):
        raise ValueError("Not all items were visited in the segment.")
    if len(ordered_connections) != len(connections):
        raise ValueError("Not all connections were visited in the segment.")

    return ordered_items, ordered_connections


def sort_segment_items_and_connections(the_segment: piping.PipingNetworkSegment) -> None:
    """
    Sorts the connections and items of the segment in the order of occurrence along the segment.

    The first connection and item are determined by the source and target item of the segment. The
    connections and items are then sorted in place.

    Parameters
    ----------
    the_segment : pipingNetworkSegment
        The segment to be sorted.

    Returns
    -------
    None: the_segment is manipulated in place.
    """
    # Sort connections and items in order of occurrence along the segment.
    try:
        items, connections = sort_connected_items_and_connections(
            the_segment.items, the_segment.connections
        )
    except ValueError as e:
        raise DexpiCorruptPipingSegmentException(
            f"Segment {the_segment} is not validly connected. {e}"
        )
    the_segment.connections = connections
    the_segment.items = items


def piping_network_segment_validity_check(
    the_segment: piping.PipingNetworkSegment,
) -> tuple[PipingValidityCode, str]:
    """
    Performs a consistency check on the pipingNetworkSegment.

    Returns an appropriate PipingValidityCode Enum type. Internally, this is done by going
    along all items and connections to see if they form a valid chain.

    Parameters
    ----------
    the_system : pipingNetworkSegment
        The piping network segment to be checked for consistency.

    Returns
    -------
    (PipingValidityCode, str)
        An Enum type indicating the validity status of the piping network
        segment. msg contains the corresponding error message.
    """

    visited_items = []
    visited_connections = []

    # First, go through connections to find the first connection of the segment.
    # This should have the same source spec as the PNS (at least, this is our
    # interpretation of the proteus spec. Adjust in future if otherwise)
    first_connection = None
    first_is_found = False
    for connection in the_segment.connections:
        # Find the first connection in the segment
        if the_segment.sourceItem == connection.sourceItem:
            first_connection = connection
            if not first_is_found:
                first_is_found = True
            else:
                msg = (
                    "At least two pipes in this segment have the same "
                    "piping input as the parent segment, should be one."
                )
                return (PipingValidityCode.INTERNAL_VIOLATION, msg)
    if first_connection is None and the_segment.connections:
        msg = "No piping connection found with the same piping input as the parent segment"
        return (PipingValidityCode.CONNECTION_CONVENTION_VIOLATION, msg)
    current_connection = first_connection
    visited_connections.append(current_connection)
    # Find the first item as source of the first pipe. Is none if the source
    # is outside the segment
    first_item = the_segment.sourceItem if the_segment.sourceItem in the_segment.items else None
    if first_item is not None:
        visited_items.append(first_item)
        current_item = first_item

    # Iterate along the pns one connection and item at a time to find all
    # connected segment members
    while True:
        # Find current connections destination
        current_item = current_connection.targetItem
        # Check if the end of the segment is reached, and finish. The
        # destination could be none, which is a valid case only at the end of
        # the segment
        if current_item is None and the_segment.targetItem is not None:
            msg = f"No destination to pipe {current_connection} referenced "
            return (PipingValidityCode.INTERNAL_VIOLATION, msg)
        if current_item in visited_items:
            msg = f"Item {current_item} referenced as destination twice inone segment"
            return (PipingValidityCode.INTERNAL_VIOLATION, msg)
        if current_item is not None:
            visited_items.append(current_item)
        if the_segment.targetItem == current_item:
            break
        # Otherwise, find the next connection. Return internal violation if none
        # found
        for connection in the_segment.connections:
            connection_found = False
            if connection.sourceItem == current_item:
                current_connection = connection
                visited_connections.append(current_connection)
                connection_found = True
                break
        if not connection_found:
            if len(visited_connections) != len(the_segment.connections):
                msg = (
                    "Piping network segment is interrupted, last connected "
                    "pipe does and not share the same destination as the segment and not all connections are visited"
                )
                return (PipingValidityCode.INTERNAL_VIOLATION, msg)
            else:
                msg = (
                    "Final pipe does not share its destination item with the piping network segment"
                )
                return (PipingValidityCode.CONNECTION_CONVENTION_VIOLATION, msg)

    # Check if there are connections or pipes in the segment that werent
    # visited, in which case it is corrupt
    if len(visited_connections) < len(the_segment.connections):
        msg = "Not all connections in the segment were visited"
        return (PipingValidityCode.INTERNAL_VIOLATION, msg)
    if len(visited_items) < len(the_segment.items):
        msg = "Not all items in the segment were visited"
        return (PipingValidityCode.INTERNAL_VIOLATION, msg)

    # Check if the order of the connections and items is the order of occurrence
    for segment_connection, visited_connection in zip(the_segment.connections, visited_connections):
        if segment_connection != visited_connection:
            msg = "Segment connections not in right order"
            return (PipingValidityCode.INTERNAL_VIOLATION, msg)
    for segment_item, visited_item in zip(the_segment.items, visited_items):
        if segment_item != visited_item:
            msg = "Segment item not in right order"
            return (PipingValidityCode.INTERNAL_VIOLATION, msg)

    # Investigate if all nodes referenced are also part of the respective item
    elements_to_examine = [the_segment]
    elements_to_examine.extend(visited_connections)
    for element in elements_to_examine:
        for item, node, type_str in zip(
            [element.sourceItem, element.targetItem],
            [element.sourceNode, element.targetNode],
            ["source", "target"],
        ):
            if item is None and node is not None:
                msg = f"{element} has a {type_str} node but no {type_str} item"
                return (PipingValidityCode.INTERNAL_VIOLATION, msg)
            if item is not None and node is not None:
                if node not in item.nodes:
                    msg = f"{element} {type_str} node not a member of its {type_str} item"
                    return (PipingValidityCode.INTERNAL_VIOLATION, msg)

    # If no violation was encountered above, the segment is valid
    return (PipingValidityCode.VALID, "Segment valid")


def _connect_piping_connection(
    connector_item: piping.PipingSourceItem | piping.PipingTargetItem,
    piping_connection: piping.PipingConnection | None = None,
    piping_segment: piping.PipingNetworkSegment | None = None,
    connector_node: piping.PipingNode | None = None,
    as_source: bool = False,
    force_reconnect: bool = False,
) -> None:
    """
    Connects the pipe to the target connector object.

    If the connector object is a PipingNode, the connection is implemented via
    node connection. Otherwise, a connection via Source or TargetItem is
    attempted. Raises an error if the piping connection is already connected
    unless `force_reconnect` is True.

    Parameters
    ----------
    connector_item : Union[PipingSourceItem, PipingTargetItem]
        The item to which the pipe will be connected.
    piping_connection : PipingConnection, optional
        The piping connection to be reconnected. If not passed, then
        reconnection only performed on the segment
    piping_segment : PipingNetworkSegment, optional
        Applies the reconnection to the piping segment too if passed along.
    connector_node: PipingNode, optional
        Optionally, the node of connector_item to be connected to. Must be a
        member of connector_item
    as_source : bool, optional
        If True, the connection is established as source, else as target.
        Defaults to False.
    force_reconnect : bool, optional
        If False, raises an exception if the piping connection is already
        connected. Defaults to False.

    Returns
    -------
    None
        This function modifies the state of the piping connection but does
        not return any value.

    Raises
    ------
    ConnectionException
        If the piping connection is already connected, and `force_reconnect`
        is False.
    ValueError
        If the piping_connection and piping_segment are found to be faulty or
        do not fit together, or if the optional connector_node is not associated to the connector_item
    """
    if connector_item is None and connector_node is not None:
        msg = "Cannot have a node without an item for connections."
        raise ValueError(msg)
    if connector_node is not None:
        if connector_node not in connector_item.nodes:
            msg = f"Piping node {connector_node} not associated to item {connector_item}"
            raise ValueError(msg)

    def raise_exception_if_connection_conflict(connection_object, as_source) -> None:
        """Raise a DexpiConnectionException if the object is already connected"""
        if as_source and (
            connection_object.sourceNode is not None or connection_object.sourceItem is not None
        ):
            raise DexpiConnectionException(
                f"{type(connection_object)} object: {connection_object} already has a source."
            )
        elif not as_source and (
            connection_object.targetNode is not None or connection_object.targetItem is not None
        ):
            raise DexpiConnectionException(
                f"{type(connection_object)} object: {connection_object} already has a target."
            )

    def reconnect_connection_object(conntection_object, item, node, as_source) -> None:
        """Set the source/target connector to the item and node specified"""
        if as_source:
            conntection_object.sourceNode = None
            conntection_object.sourceItem = None
            conntection_object.sourceItem = item
            if connector_node is not None:
                conntection_object.sourceNode = node

        else:
            conntection_object.targetNode = None
            conntection_object.targetItem = None
            conntection_object.targetItem = item
            if connector_node is not None:
                conntection_object.targetNode = node

    if piping_segment is not None and piping_connection is not None:
        final_segment_connection = find_final_connection(piping_segment, as_source=as_source)
        if final_segment_connection != piping_connection:
            msg = f"{piping_connection} is not the final connection of {piping_segment}"
            raise ValueError(msg)

    # check if connection already exists.
    if not force_reconnect:
        if piping_connection is not None:
            raise_exception_if_connection_conflict(piping_connection, as_source)
        if piping_segment is not None:
            raise_exception_if_connection_conflict(piping_segment, as_source)

    # Perform reconnection
    if piping_connection is not None:
        reconnect_connection_object(piping_connection, connector_item, connector_node, as_source)
    if piping_segment is not None:
        reconnect_connection_object(piping_segment, connector_item, connector_node, as_source)



"""This module contains the class for the SyntheticPIDGenerator, which
generates synthetic P&IDs.

For this, P&ID patterns as defined in pattern.py are connected via their
connector interfaces, with a logic predefined in a given generator function.
"""

from pydexpi.syndata.connector_renaming import ConnectorRenamingConvention
from pydexpi.syndata.generation_history import GenerationHistory
from pydexpi.syndata.generator_function import CappingFunction, GeneratorFunction
from pydexpi.syndata.pattern import Pattern


class SyntheticPIDGenerator:
    """The generator contains mechanisms to connect patterns to a larger pattern
    and collect them in pattern distributions.

    Attributes
    ----------
    _current_pattern: Pattern
        This attribute stores the (partially) generated P&ID as a pattern.
    generator_function: GeneratorFunction
        The generator function that defines the generation strategy to be
        applied by this generator.
    capping_function: CappingFunction
        The capping function that defines the capping strategy to be applied by this generator.
    renaming_convention: ConnectorRenamingConvention
        The renaming convention that renames connectors of new patterns that are included.
    max_steps: int
        The maximum number of steps to take before early termination.
    _generation_history: GenerationHistory
        The generation history of the generator, which stores the steps taken
        during the generation process.
    _current_step: int
        Attribute keeping track of the current step the generator is in.
    _termination_flag: bool
        Internal flag that can be used to invoke termination of the current
        generation process.
    """

    def __init__(
        self,
        generator_function: GeneratorFunction,
        capping_function: CappingFunction = None,
        renaming_convention: ConnectorRenamingConvention = None,
        max_steps=100,
    ) -> None:
        """Constructor of a synthetic P&ID generator

        Parameters
        ----------
        generator_function : GeneratorFunction
            The function that defines the generation logic to be followed
        capping_function : CappingFunction, optional
            The function that defines the capping logic to be followed, by default None
        renaming_convention : ConnectorRenamingConvention, optional
            The function that defines the renaming logic to be followed, by default None
        max_steps : int, optional
            The maximum amount of steps to be taken before terminating the generation,
            by default 100
        """
        self._current_pattern: Pattern = None
        self.generator_function = generator_function
        self.capping_function = capping_function
        self.renaming_convention = renaming_convention or ConnectorRenamingConvention()
        if max_steps <= 0:
            raise ValueError("max_steps must be a positive integer.")
        self.max_steps = max_steps
        self._generation_history = GenerationHistory()
        self._current_step = 0
        self._termination_flag = False

    def generate_pattern(self, new_pattern_label: str) -> Pattern:
        """Top-level function to create a random dexpi flowsheet.

        For this, a pattern is first initialized according to the initialization
        defined by the generator_function. Then, while no stopping criteria are
        reached, the _current_pattern is aggregated iteratively with the
        next_step() function. Finally, unconnected connectors are dropped by
        calling drop_connector() on all connectors of the pattern, and the
        generated pattern is returned. The _current_pattern is reset to None
        for the next call of generate_pattern().

        Parameters
        ----------
        new_pattern_label : str
            The label the new pattern should have.

        Returns
        -------
        Pattern:
            The generated pattern.

        Raises
        ------
        AttributeError:
            If the _current_pattern was already initialized before the call of
            this method.
        """
        if self.max_steps <= 0:
            return

        # First, reset the SyntheticPIDGenerator
        self.reset()

        # Use new interface to initialize the pattern
        init_step = self.generator_function.initialize_pattern()
        self._current_step += 1
        self._generation_history.write_step(init_step)
        self._current_pattern = init_step.get_pattern()

        # Rename the new pattern before proceeding
        self._current_pattern.change_label(new_pattern_label)

        # Generate pattern as long as no stopping criterion are reached
        while self.continue_loop():
            self.next_step()
            self._current_step += 1

        # Cap off connectors according to the capping function
        if self.capping_function is not None:
            self.cap_connectors()

        return self._current_pattern

    def next_step(self) -> None:
        """Adds a further pattern to the P&ID, or performs an internal
        connection, by querying the generator function for the next step.

        Raises
        ------
        AttributeError:
            If the _current_pattern is not yet initialized and therefore None.
        """
        if self._current_pattern is None:
            msg = "Pattern not yet initialized"
            raise AttributeError(msg)

        # Get next step from generator function
        step = self.generator_function.get_next_step(self._current_pattern)

        # Write step to the generation history (Do this before renaming to preserve the state
        # at which the step was created by the generator function (Shouldnt make a difference
        # atm, but may make a difference in the future))
        self._generation_history.write_step(step)

        # Apply renaming, execute step, and check termination status
        step.apply_renaming_convention(self.renaming_convention)
        step.execute_on(self._current_pattern)
        self._termination_flag = step.get_termination_status()

    def continue_loop(self) -> bool:
        """Checks stopping criteria for P&ID generation. Stops if no open
        connectors are available or the maximum number of generation steps is
        reached.

        Returns
        -------
        bool:
            True if none of the stopping criteria are reached, False otherwise.
        """
        if (
            self._current_pattern.connectors
            and self._current_step < self.max_steps
            and not self._termination_flag
        ):
            return True
        else:
            return False

    def cap_connectors(self) -> None:
        """Cap the connectors of the current pattern using the capping function."""
        if self._current_pattern is None:
            msg = "Pattern not yet initialized"
            raise AttributeError(msg)

        capping_steps = self.capping_function.get_capping_steps(self._current_pattern)
        for step in capping_steps:
            self._generation_history.write_step(step)
            step.execute_on(self._current_pattern)

    def get_generation_history(self) -> GenerationHistory:
        """Get the generation history of the generator.

        Returns
        -------
        GenerationHistory:
            The generation history of the generator.
        """
        return self._generation_history

    def reset(self) -> None:
        """Reset the attributes pattern, step counter, and termination flag
        to prepare generator to generate a new pattern.
        """
        self._current_pattern = None
        self._termination_flag = False
        self._current_step = 0

        # Reset the renaming convention and make a new generation history
        self.renaming_convention.reset()
        self._generation_history = GenerationHistory()



"""This module contains child classes for pattern and connectors to use
the synthetic data generator on GraphML graphs."""

from __future__ import annotations

import uuid

import networkx as nx

from pydexpi.syndata.pattern import Connector, Pattern


class GraphConnector(Connector):
    """Extends the Connector for a P&ID graph representation in networkx.

    Defines common attributes for graph connectors for graphs with unattributed
    edges. Edge attributes can be added in subclass implementations by
    overriding `get_edge_attrs()`. This design enables adding type
    information to edge attributes while maintaining extensibility for graph
    connection types.

    A graph connector can only connect to another connector of the same class
    with an opposite inlet flag. Subclasses provide a `get_edge_type()` method
    to determine specific edge types.

    Because NetworkX Stores connections on a graph level rather than a node
    level, the graph connector does not contain any connection functionality.
    Connection logic is performed in pattern with _connect_via_pattern.

    Inherited Attributes
    -------------------
    label : str
        The label for the connector, inherited from `Connector`.
    kwinfos : dict, optional
        Additional keyword arguments for the connector, inherited from
        `Connector`.

    Attributes
    ----------
    reference_node_id : str
        The id of the node referenced as connection interface
    is_inlet : bool
        If this connector is a source or a destination for the directed edge
    """

    def __init__(
        self,
        label: str,
        reference_node_id: str,
        is_inlet: bool,
        kwinfos: dict = None,
    ) -> None:
        """
        Initializes a GraphConnector.

        Parameters
        ----------
        label : str
            The label for the connector.
            reference_node_id : str
        The id of the node referenced as connection interface
        connector_type : ConnectorType
            The type of the connector, used to ensure valid connection
            combinations.
        kwinfos : dict, optional
            Additional keyword arguments for the connector.
        """
        super().__init__(label, kwinfos)
        self.reference_node_id = reference_node_id
        self.is_inlet = is_inlet

    def get_edge_attrs(self) -> dict:
        """Returns the connection type to be placed on the edge attributes of
        the graph. Implementation is specific to subclasses. Recommended
        implementation via static, class level dict attribute.

        Returns
        -------
        dict:
            The attribute kw args to be placed on any new edge."""
        return {}

    def assess_valid_counterpart(self, counterpart: GraphConnector) -> bool:
        """Assesses the validity of a counterpart graph connector. A graph
        connector is valid if it is of the same class and has the opposite
        is_inlet flag.

        Parameters
        ----------
        counterpart: GraphConnector
            The counterpart to be checked for validity.

        Returns
        -------
        bool:
            True if the counterpart is valid, False otherwise.
        """
        validity = isinstance(counterpart, self.__class__) and self.is_inlet ^ counterpart.is_inlet
        return validity

    def _implement_connection(self, counterpart: GraphConnector) -> None:
        """Empty implementation of _implement_connection. The implementation
        is empty, because connection is performed on pattern level via
        _connect_via_pattern."""
        pass


class GraphBasicPipingConnector(GraphConnector):
    """Simple implementation for attributed pipe edges.
    See GraphConnector for general behavior."""

    edge_type_key = "edge_type"
    edge_type = "PipingConnection"

    @classmethod
    def get_edge_attrs(cls) -> dict:
        return {cls.edge_type_key: cls.edge_type}


class GraphBasicSignalConnector(GraphConnector):
    """Simple implementation for attributed signal edges.
    See GraphConnector for general behavior."""

    edge_type_key = "edge_type"
    edge_type = "SignalConnection"

    @classmethod
    def get_edge_attrs(cls) -> dict:
        return {cls.edge_type_key: cls.edge_type}


class GraphPattern(Pattern):
    """
    Extends the Pattern class to represent a graph pattern. This class wrapps a
    networkx graph and extends functionality for connecting patterns and
    incorporating other graph patterns. To ensure unique hashability, each node
    is assigned a uuid as label, with any associated data to be placed in node
    data.

    Inherited Attributes
    ----------
    label : str
        The label for the pattern.
    connectors : list[Connector]
        A list of connectors associated with this pattern.
    observer_patterns : dict[str, Pattern], optional
        A dictionary of observer Pattern objects. If None, no observer
        patterns are used.
    kwinfos : dict, optional
        Additional keyword information for the pattern.

    Attributes
    ----------
    the_graph : nx.DiGraph
        The networkx graph associated with this pattern.
    """

    def __init__(
        self,
        label: str,
        the_graph: nx.DiGraph,
        connectors: dict[str, GraphConnector],
        observer_patterns: dict[str, Pattern] = None,
        kwinfos: dict = None,
    ) -> None:
        """
        Initializes a GraphPattern with the given label, connectors, and graph.

        Parameters
        ----------
        label : str
            The label for the pattern.
        connectors : list[Connector]
            A list of connectors associated with this pattern.
        the_graph : nx.DiGraph
        The networkx graph associated with this pattern.
        observer_patterns : dict[str, Pattern], optional
            A dictionary of observer Pattern objects. If None, no observer
            patterns are used.
        kwinfos : dict, optional
            Additional keyword information for the pattern.

        Raises
        ------
        ValueError
            If any connector reference not in the graph nodes.
        """
        # For backwards compatibility, cast a list of connectors to a dict
        if isinstance(connectors, list):
            connectors = {connector.label: connector for connector in connectors}

        # First, do some consistency checks
        for connector in connectors.values():
            if connector.reference_node_id not in the_graph.nodes:
                msg = (
                    f"The node {connector.reference_node_id} is not associated "
                    f"to graph {the_graph}. Pattern couldnt be created."
                )
                raise ValueError(msg)

        # Construct the object
        super().__init__(label, connectors, observer_patterns, kwinfos)
        self.the_graph = the_graph

    def _implement_incorporation(self, counterpart: GraphPattern) -> None:
        """Imports all nodes and edges of the counterpart into this patterns
        graph.

        Parameters
        ----------
        counterpart : GraphPattern
            The counterpart to be incorporated.

        Raises
        ------
        ValueError
            If any node or edge label of the counterpart is already found in
            this patterns graph.
        """
        # Check for overlapping nodes
        if set(self.the_graph.nodes).intersection(counterpart.the_graph.nodes):
            raise ValueError("Node overlap detected between G1 and G2.")

        # Check for overlapping edges
        if set(self.the_graph.edges).intersection(counterpart.the_graph.edges):
            raise ValueError("Edge overlap detected between G1 and G2.")

        super()._implement_incorporation(counterpart)

        # Import nodes and edges
        self.the_graph.add_nodes_from(counterpart.the_graph.nodes(data=True))
        self.the_graph.add_edges_from(counterpart.the_graph.edges(data=True))

    def _connect_via_pattern(
        self, connector: GraphConnector, counterpart_connector: GraphConnector
    ) -> None:
        """Connect the connectors by adding a new edge to this patterns graph
         object.

        Parameters
        ----------
        connector : GraphConnector
            The connector to be connected to counterpart
        counterpart_connector : GraphConnector
            The counterpart connector to be connected to the connector.
        """
        # Determine edge direction
        if counterpart_connector.is_inlet:
            source_id = connector.reference_node_id
            target_id = counterpart_connector.reference_node_id
        else:
            source_id = counterpart_connector.reference_node_id
            target_id = connector.reference_node_id

        # Determine edge label
        edge_kwargs = connector.get_edge_attrs()

        # Perform connection
        self.the_graph.add_edge(source_id, target_id, **edge_kwargs)

    def copy_pattern(self, *args, **kwargs):
        """Override parent copy method to assign a new UUID to the wrapped
        graph.

        Returns
        -------
        GraphPattern
            The deepcopied pattern
        """
        copied_pattern = super().copy_pattern(*args, **kwargs)
        new_label_map = {
            old_label: str(uuid.uuid4()) for old_label in copied_pattern.the_graph.nodes
        }
        copied_pattern.the_graph = nx.relabel_nodes(copied_pattern.the_graph, new_label_map)

        for key, connector in copied_pattern.connectors.items():
            connector.reference_node_id = new_label_map[connector.reference_node_id]
        return copied_pattern



"""Generator function that encapsulates the logic of the P&ID generation.

This module contains the abstract definition of a generator function used
by the synthetic data generator, following a strategy design pattern.
A particular logic may be defined in a child class of the generator function.
It also contains a sample implementation called RandomGeneratorFunction that
performs all actions completely at random."""

import random
from abc import ABC, abstractmethod

from pydexpi.syndata.generation_history import GenerationHistory
from pydexpi.syndata.generator_step import (
    AddPatternStep,
    CappingStep,
    GeneratorStep,
    InitializationStep,
    InternalConnectionStep,
    TerminationStep,
)
from pydexpi.syndata.pattern import (
    Pattern,
)
from pydexpi.syndata.pattern_distribution import PatternDistribution


class GeneratorFunction(ABC):
    """Abstract class for the generator function. Generator function contains
    the logic of determining the next pattern to be connected, in accordance
    with the strategy design pattern. The logic follows:

    current_pattern_connection,
    next_pattern,
    next_pattern_connection = f(current_pattern)

    A recycle can be produced by returning two connectors of the current pattern
    and no next pattern.

    Attributes
    ----------
    distribution_range: dict[str, PatternDistribution]
        This functions range of possible PatternDistributions to
        sample from.
    """

    def __init__(self, distribution_range: dict[str, PatternDistribution]) -> None:
        """Constructor for an abstract generator function class.

        Takes the distribution range as argument, which is the range of possible
        distributions this generator function samples from. The dictionary must
        have the distribution names as keys.

        Parameters
        ----------
        distribution_range : dict[str, PatternDistribution]
            The range of possible PatternDistributions to sample from.

        Raises
        ------
        ValueError
            If the names of the distributions don't match the keys of the
            distribution_range dictionary argument.
        """
        for the_name in distribution_range.keys():
            if distribution_range[the_name].name != the_name:
                msg = "Distribution names dont match range keys"
                raise ValueError(msg)

        self.distribution_range = distribution_range

    @abstractmethod
    def get_next_step(self, current_pattern: Pattern) -> GeneratorStep:
        """Dictates the next operation to be performed in the generation process.

        For this, the current pattern is passed as argument and the next
        operation is returned as a GeneratorStep.
        """

    @abstractmethod
    def initialize_pattern(self) -> InitializationStep:
        """Initializes the P&ID generation with a seed pattern.

        Returns
        -------
        InitializationStep
            Initialization step containing the seed pattern and distribution name.
        """


class CappingFunction:
    """Abstract class for the capping function.

    Capping function is used to define how to cap the pattern's remaining, unconnected connectors at
    the end of the generation process. Default implementation is to cap all connectors with a
    default capping step.

    Attributes
    ----------
    distribution_range: dict[str, PatternDistribution]|None
        The range of possible PatternDistributions to sample from. Can be None if not used.

    """

    def __init__(self, distribution_range: dict[str, PatternDistribution] | None = None) -> None:
        """Constructor for the capping function.

        Takes the distribution range as argument, which is the range of possible
        distributions this generator function samples from. The dictionary must
        have the distribution names as keys.

        Parameters
        ----------
        distribution_range : dict[str, PatternDistribution] | None, optional
            The range of possible PatternDistributions to sample from, by default None.
        """
        if distribution_range:
            for distribution in distribution_range.values():
                if not len(distribution.connector_labels) == 1:
                    msg = "CappingFunction should only use distributions with one connector label"
                    raise ValueError(msg)

        self.distribution_range = distribution_range

    def get_capping_steps(self, current_pattern: Pattern) -> list[CappingStep]:
        """Returns a list of capping steps for the given pattern.

        Retrieves the capping steps from the subclass-specific method _make_capping_steps, and
        performs some consistency checks on the result.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to be capped.

        Returns
        -------
        list[GeneratorStep]
            A list of capping steps, one for each connector in the current pattern.

        Raises
        ------
        RuntimeError
            If not all connectors are capped in the capping steps.
        """
        capping_steps = self._make_capping_steps(current_pattern)
        if {step.own_connector for step in capping_steps} != set(
            current_pattern.connectors.values()
        ):
            raise RuntimeError("Connectors of the CappingSteps do not match the current pattern.")
        return capping_steps

    def _make_capping_steps(self, current_pattern: Pattern) -> list[CappingStep]:
        """Generates capping steps for the given pattern.

        This method produces the CappingSteps for the given pattern. It is intended to be overridden
        by subclasses with more complex capping mechanisms. By default, it creates a default capping
        step for each connector (Without incorporating any further elements).

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to be capped.

        Returns
        -------
        list[GeneratorStep]
            A list of capping steps, one for each connector in the current pattern.
        """
        return [CappingStep(connector) for connector in current_pattern.connectors.values()]


class RandomGeneratorFunction(GeneratorFunction):
    """Generator function that selects next patterns and connectors fully at
    random.

    Inherited attributes
    ----------
    distribution_range: dict[str, PatternDistribution]
        This functions range of possible PatternDistributions to
        sample from.

    Attributes
    ----------
    p_connect_internal: float
        The probability to attempt to connect a recycle at each step. If no
        valid recycle connections are found, a new pattern is sampled instead.
    use_distribution_probabilities: bool
        If false, the patterns are sampled from the distribution at random
        instead of using the distribution probabilities.
    """

    def __init__(
        self,
        distribution_range: dict[str, PatternDistribution],
        p_connect_internal: float = 0.0,
        use_distribution_probabilities: bool = True,
    ) -> None:
        """Constructor for a generator function class that selects patterns
        and connectors completely at random.

        Takes the distribution range as argument, which is the range of possible
        distributions this generator function samples from. The dictionary must
        have the distribution names as keys.

        Parameters
        ----------
        distribution_range : dict[str, PatternDistribution]
            The range of possible PatternDistributions to sample from.
        p_connect_internal : float
            The probability that a recycle connection is attempted at each step.
            Must be between 0 and 1.
        use_distribution_probabilities : bool, optional
            Whether to use the pattern distribution probabilities while
            sampling. If False, patterns are sampled fully at random.
            By default True

        Raises
        ------
        ValueError
            If the names of the distributions don't match the keys of the
            distribution_range dictionary argument.

        """
        super().__init__(distribution_range)
        self.p_connect_internal = None
        self.set_p_connect_internal(p_connect_internal)
        self.use_distr_probs = use_distribution_probabilities

    def get_next_step(self, current_pattern: Pattern) -> GeneratorStep:
        """Implements the generator function to decide the next operation.

        First, a decision is made whether to connect a recycle or sample a new
        pattern. If a recycle is to be connected, a random internal connection is obtained from
        select_internal_connector. If no valid internal connection is found or the decision is
        made to sample a new pattern, the select_next_pattern function is called.

        Parameters
        ----------
        current_pattern : Pattern
            Current state of the P&ID.

        Returns
        -------
        GeneratorStep
            The next step to be performed
        """
        connect_internal = random.choices(
            [True, False],
            weights=[self.p_connect_internal, 1 - self.p_connect_internal],
        )[0]
        if connect_internal:
            # Select random internal connectors
            internal_connection_step = self.select_internal_connector(current_pattern)
            # If no compatible, internal connectors are found, default to sampling a new pattern
            if internal_connection_step is None:
                return self.select_next_pattern(current_pattern)
            else:
                return internal_connection_step
        else:
            return self.select_next_pattern(current_pattern)

    def initialize_pattern(self) -> InitializationStep:
        """Initializes the P&ID generation with a seed pattern that is selected
        at random.

        Returns
        -------
        InitializationStep
            Initialization step containing the seed pattern and distribution name.
        """
        # Randomly choose a distribution
        next_distribution = random.choice(list(self.distribution_range.values()))

        # Randomly choose a pattern
        init_pattern, _ = (
            next_distribution.sample_pattern()
            if self.use_distr_probs
            else next_distribution.random_pattern()
        )

        return InitializationStep(
            init_pattern=init_pattern,
            sampled_distribution_name=next_distribution.name,
        )

    def select_next_pattern(self, current_pattern: Pattern) -> AddPatternStep | TerminationStep:
        """Selects the next pattern and connectors at random.

        For this, a random connector is selected in the current pattern. Then,
        a random pattern distribution is selected and a pattern sampled from it.
        A suitable connector is selected at random in the next pattern, and the
        sampled connectors and pattern are returned. If no suitable connector is
        available in the patterns of the selected distribution, a different
        distribution is selected randomly. If no distributions fit, a different
        own_connector is selected for the next step, and the connector is
        labeled as unconnectable for future steps.

        If no further connectable connectors remain, return None, None, None

        Parameters
        ----------
        current_pattern : Pattern
            Current state of the P&ID.

        Returns
        -------
        AddPatternStep|TerminationStep
            The next step to be performed. If no further connectable connectors are found,
            a TerminationStep is returned. Otherwise, an AddPatternStep is returned.
        """
        visited_connectors = set()
        while visited_connectors != set(current_pattern.connectors):
            own_conn_choices = [
                conn for conn in current_pattern.connectors if conn not in visited_connectors
            ]
            if not own_conn_choices:
                break

            own_connector_key = random.choice(own_conn_choices)
            visited_connectors.add(own_connector_key)
            own_connector = current_pattern.connectors[own_connector_key]

            if "Connectable" in own_connector.kwinfos:
                if not own_connector.kwinfos["Connectable"]:
                    continue

            pattern_found = False
            visited_distributions = set()
            while visited_distributions != set(self.distribution_range.values()):
                distr_choices = [
                    distr
                    for distr in self.distribution_range.values()
                    if distr not in visited_distributions
                ]

                next_distribution = random.choice(distr_choices)
                visited_distributions.add(next_distribution)

                next_pattern, _ = (
                    next_distribution.sample_pattern()
                    if self.use_distr_probs
                    else next_distribution.random_pattern()
                )

                next_conn_choices = [
                    label
                    for label, conn in next_pattern.connectors.items()
                    if own_connector.assess_valid_counterpart(conn)
                ]
                if next_conn_choices:
                    next_connector_key = random.choice(next_conn_choices)
                    next_connector = next_pattern.connectors[next_connector_key]
                    next_step = AddPatternStep(
                        own_connector=own_connector,
                        next_pattern=next_pattern,
                        next_connector=next_connector,
                        sampled_distribution_name=next_distribution.name,
                    )
                    pattern_found = True
                    return next_step

            if not pattern_found:
                own_connector.kwinfos["Connectable"] = False

        return TerminationStep()

    def select_internal_connector(self, current_pattern: Pattern) -> InternalConnectionStep | None:
        """Selects the internal connectors for internal connector at random.

        For this, a random connector is selected in the current pattern. Then
        a second connector is selected randomly from valid counterparts in the
        current pattern. If no valid counterparts are found, a different
        first_connector is selected for the next step.

        If no connector pairs are found in current pattern, return None, None

        Parameters
        ----------
        current_pattern : Pattern
            Current state of the P&ID.

        Returns
        -------
        InternalConnectionStep|None
            The next step to be performed. If no further connectable connectors are found,
            None is returned.

        """
        visited_connectors = set()
        while visited_connectors != set(current_pattern.connectors):
            first_conn_choices = [
                conn for conn in current_pattern.connectors if conn not in visited_connectors
            ]

            first_connector_key = random.choice(first_conn_choices)
            visited_connectors.add(first_connector_key)
            first_connector = current_pattern.connectors[first_connector_key]

            if "Connectable" in first_connector.kwinfos:
                if not first_connector.kwinfos["Connectable"]:
                    continue

            scnd_conn_choices = [
                label
                for label, conn in current_pattern.connectors.items()
                if first_connector.assess_valid_counterpart(conn)
                and conn not in visited_connectors
                and conn != first_connector
            ]
            if scnd_conn_choices:
                scnd_connector_key = random.choice(scnd_conn_choices)
                scnd_connector = current_pattern.connectors[scnd_connector_key]
                return InternalConnectionStep(
                    own_connector=first_connector,
                    next_connector=scnd_connector,
                )

        return None

    def set_p_connect_internal(self, p_connect_internal: float) -> None:
        """
        Sets the internal connection probability. Raises ValueError if the
        probability is out of the bounds [0,1].

        Parameters
        ----------
        p_connect_internal : float
            The new probability of internal connection, between 0 and 1 (incl).

        Raises
        ------
        ValueError
            If `p_connect_internal` is not within the range [0, 1].
        """
        if not 0 <= p_connect_internal <= 1:
            msg = "Parameter p_connect_internal needs to be between 0 and 1."
            raise ValueError(msg)
        self.p_connect_internal = p_connect_internal


class ReconstructionGeneratorFunction(GeneratorFunction):
    """Generator function that reconstructs a data point from a generation history object."""

    def __init__(
        self,
        distribution_range: dict[str, PatternDistribution],
        generation_history: GenerationHistory | None = None,
    ) -> None:
        """Constructor for the reconstruction generator function.

        Parameters
        ----------
        distribution_range : dict[str, PatternDistribution]
            The range of possible PatternDistributions to sample from.
        """
        super().__init__(distribution_range)
        self.generation_history = generation_history
        self._current_step = 0

    def get_next_step(self, current_pattern: Pattern) -> GeneratorStep:
        """Implements the generator function to decide the next operation based on the history.

        The next step is determined by looking at the generation history and
        reconstructing the steps that were taken to generate the pattern.

        Parameters
        ----------
        current_pattern : Pattern
            Current state of the synthetic data point.

        Returns
        -------
        GeneratorStep
            The next step to be performed.

        Raises
        ------
        AttributeError
            If the generation history is not set.
        RuntimeError
            If an unexpected step sequence is encountered or if the history does not match the
            current state or the specified distribution range.
        """
        if self.generation_history is None:
            raise AttributeError("Generation history is not set.")
        if self._current_step == 0:
            raise RuntimeError("Initialization step not yet performed, step counter still at 0")

        # Get the last step from the generation history
        last_step = self.generation_history.history[self._current_step]
        sampled_distribution_name = last_step["sampled_distribution_name"]
        own_connector_label = last_step["own_connector"]
        next_pattern_label = last_step["next_pattern"]
        next_connector_label = last_step["next_connector"]

        # Create the appropriate step based on the last step type
        match last_step["generator_step_type"]:
            case "initialization":
                raise RuntimeError("Cannot encounter an initialization step after the first step.")
            case "add_pattern":
                # Retrieve the pattern
                try:
                    the_distribution = self.distribution_range[sampled_distribution_name]
                except KeyError:
                    raise RuntimeError(f"Unknown distribution name: {sampled_distribution_name}")
                try:
                    next_pattern = the_distribution.patterns[next_pattern_label]
                except KeyError:
                    raise RuntimeError(f"Unknown pattern label: {next_pattern_label}")
                # Retrieve the connectors
                try:
                    own_connector = current_pattern.connectors[own_connector_label]
                except KeyError:
                    raise RuntimeError(
                        f"Unknown connector label in current_pattern: {own_connector_label}"
                    )
                try:
                    next_connector = next_pattern.connectors[next_connector_label]
                except KeyError:
                    raise RuntimeError(
                        f"Unknown connector label in next_pattern: {next_connector_label}"
                    )
                the_step = AddPatternStep(
                    own_connector=own_connector,
                    next_pattern=next_pattern,
                    next_connector=next_connector,
                    sampled_distribution_name=sampled_distribution_name,
                )
            case "internal_connection":
                # Retrieve the connectors
                try:
                    own_connector = current_pattern.connectors[own_connector_label]
                except KeyError:
                    raise RuntimeError(
                        f"Unknown connector label in current_pattern: {own_connector_label}"
                    )
                try:
                    next_connector = current_pattern.connectors[next_connector_label]
                except KeyError:
                    raise RuntimeError(
                        f"Unknown connector label in current_pattern: {next_connector_label}"
                    )
                the_step = InternalConnectionStep(
                    own_connector=own_connector,
                    next_connector=next_connector,
                )
            case "termination":
                the_step = TerminationStep()
            case "capping":
                raise RuntimeError("Cannot encounter a capping step before Termination.")
            case _:
                raise ValueError(f"Unknown step type: {last_step['generator_step_type']}")

        # Increment the step counter and return the step
        self._current_step += 1
        return the_step

    def initialize_pattern(self) -> InitializationStep:
        """Initializes the P&ID generation with the pattern at the beginning of the history.

        Returns
        -------
        InitializationStep
            Initialization step containing the seed pattern and distribution name.
        """

        if self.generation_history is None:
            raise AttributeError("Generation history is not set.")
        if self._current_step > 0:
            msg = "Cannot get initialization step after the first step. Reset function first."
            raise RuntimeError(msg)

        # Get the first step from the generation history
        first_step = self.generation_history.history[0]
        # Create the initialization step
        try:
            the_distribution = self.distribution_range[first_step["sampled_distribution_name"]]
        except KeyError:
            raise RuntimeError(
                f"Unknown distribution name: {first_step['sampled_distribution_name']}"
            )
        try:
            init_pattern = the_distribution.patterns[first_step["next_pattern"]]
        except KeyError:
            raise RuntimeError(f"Unknown pattern label: {first_step['next_pattern']}")
        # Create the initialization step
        the_step = InitializationStep(
            init_pattern=init_pattern,
            sampled_distribution_name=first_step["sampled_distribution_name"],
        )
        # Increment the step counter
        self._current_step += 1
        return the_step

    def set_generation_history(self, generation_history: GenerationHistory) -> None:
        """Set the generation history for the reconstruction generator function.

        Parameters
        ----------
        generation_history : GenerationHistory
            The generation history to set.
        """
        self.generation_history = generation_history
        self.reset()

    def reset(self) -> None:
        """Reset the generator function counter."""
        self._current_step = 0


class ReconstructionCappingFunction(CappingFunction):
    """Capping function for the reconstruction generator function.

    This class is used to cap the connectors of a pattern based on the generation history.
    It inherits from the CappingFunction class and overrides the _make_capping_steps method.

    Attributes
    ----------
    distribution_range: dict[str, PatternDistribution]|None
        The range of possible PatternDistributions to sample from. Can be None if not used.
    """

    def __init__(
        self,
        distribution_range: dict[str, PatternDistribution],
        generation_history: GenerationHistory | None = None,
    ) -> None:
        """Constructor for the reconstruction generator function.

        Parameters
        ----------
        distribution_range : dict[str, PatternDistribution]
            The range of possible PatternDistributions to sample from.
        """
        super().__init__(distribution_range)
        self.generation_history = generation_history

    def _make_capping_steps(self, current_pattern: Pattern) -> list[CappingStep]:
        """Generates capping steps for the given pattern based on the generation history.

        This method produces the CappingSteps for the given pattern. It is intended to be overridden
        by subclasses with more complex capping mechanisms. By default, it creates a default capping
        step for each connector (Without incorporating any further elements).

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to be capped.

        Returns
        -------
        list[GeneratorStep]
            A list of capping steps, one for each connector in the current pattern.
        """
        capping_steps = []
        for i in self.generation_history:
            if i["generator_step_type"] != "capping":
                pattern_distribution_name = i["sampled_distribution_name"]
                own_connector_label = i["own_connector"]
                next_pattern = i["next_pattern"]
                next_connector_label = i["next_connector"]

                # Case: Capping step involves a pattern incorporation
                if pattern_distribution_name is None:
                    if next_pattern is not None or next_connector_label is not None:
                        raise RuntimeError(
                            "Corrupt capping step encountered: next_pattern or "
                            "next_connector should be None if pattern_distribution_name is None."
                        )
                    else:
                        the_distribution = self.distribution_range[i["sampled_distribution_name"]]
                        try:
                            next_pattern = the_distribution.patterns[next_pattern]
                        except KeyError:
                            raise RuntimeError(f"Unknown pattern label: {next_pattern}")
                        try:
                            own_connector = current_pattern.connectors[own_connector_label]
                        except KeyError:
                            raise RuntimeError(
                                f"Unknown connector label in current_pattern: {own_connector_label}"
                            )
                        try:
                            next_connector = next_pattern.connectors[next_connector_label]
                        except KeyError:
                            raise RuntimeError(
                                f"Unknown connector label in next_pattern: {next_connector_label}"
                            )

                        the_step = CappingStep(
                            own_connector=own_connector,
                            next_pattern=next_pattern,
                            next_connector=next_connector,
                            sampled_distribution_name=pattern_distribution_name,
                        )

                # Case: Capping step only involves dropping the connector
                else:
                    try:
                        own_connector = current_pattern.connectors[own_connector_label]
                    except KeyError:
                        raise RuntimeError(
                            f"Unknown connector label in current_pattern: {own_connector_label}"
                        )
                    the_step = CappingStep(
                        own_connector=own_connector,
                    )
                # Append this connectors capping step
                capping_steps.append(the_step)

        # Check if all connectors have been assigned a capping step
        if len(capping_steps) != len(current_pattern.connectors):
            raise RuntimeError("Not all connectors have a capping step in the history.")

        return capping_steps

    def set_generation_history(self, generation_history: GenerationHistory) -> None:
        """Set the generation history for the reconstruction capping function.

        Parameters
        ----------
        generation_history : GenerationHistory
            The generation history to set.
        """
        self.generation_history = generation_history



from .pydantic_classes import (
    Enterprise,
    IndustrialComplex,
    IndustrialComplexParentStructure,
    PlantArea,
    PlantAreaLocatedStructure,
    PlantSection,
    PlantSectionParentStructure,
    PlantStructureItem,
    PlantSystem,
    PlantSystemLocatedStructure,
    PlantTrain,
    PlantTrainLocatedStructure,
    ProcessPlant,
    ProcessPlantParentStructure,
    Site,
    TechnicalItem,
    TechnicalItemParentStructure,
)



"""This module contains child classes for pattern and connectors to use
the synthetic data generator on DEXPI data."""

from __future__ import annotations

from pydexpi.dexpi_classes import piping
from pydexpi.dexpi_classes.dexpiModel import DexpiModel
from pydexpi.syndata.pattern import Connector, Pattern
from pydexpi.toolkits import model_toolkit as mt, piping_toolkit as pt


class BasicPipingInConnector(Connector):
    """Extends the Connector for a basic piping in connection in DEXPI. For
    this, it wraps a PipingTargetItem and optionally an associated node.

    Inherited Attributes
    -------------------
    label : str
        The label for the connector, inherited from `Connector`.
    kwinfos : dict, optional
        Additional keyword arguments for the connector, inherited from
        `Connector`.

    Attributes
    ----------
    target_item : piping.PipingTargetItem
        The PipingTargetItem wrapped by the connection.
    target_node_index : int, optional
        The index of the target node in the PipingTargetItem. If None, no
        specific node is targeted and connections are established only via the
        item."""

    def __init__(
        self,
        label: str,
        target_item: piping.PipingTargetItem,
        target_node_index: int = None,
        kwinfos: dict = None,
    ) -> None:
        """
        Initializes a BasicPipingInConnector.

        Parameters
        ----------
        label : str
            The label for the connector.
        target_item : piping.PipingTargetItem
            The PipingTargetItem associated with the connection.
        target_node_index : int, optional
            The index of the target node in the PipingTargetItem. If None, no
            specific node is targeted.
        kwinfos : dict, optional
            Additional keyword arguments for the connector.

        Raises
        ------
        TypeError
            If `target_item` is not an instance of PipingTargetItem.
        IndexError
            If `target_node_index` is specified but is out of the range of the
            target item's nodes.
        """
        if not isinstance(target_item, piping.PipingTargetItem):
            raise TypeError("Reference object must be a DEPXI PipingTargetItem")
        if target_node_index is not None:
            if target_node_index >= len(target_item.nodes):
                raise IndexError("Target node index is out of range")
        super().__init__(label, kwinfos)
        self.target_item = target_item
        self.target_node_index = target_node_index

    def assess_valid_counterpart(self, counterpart: Connector) -> bool:
        """Implement abstract method to check for counterpart compatibility.
        Counterpart is valid if it is a BasicPipingOutConnector

        Parameters
        ----------
        counterpart : Connector
            The candidate counterpart to be assessed for connection
            compatibility.

        Returns
        -------
        bool
            True if counterpart is suitable for connection, False otherwise
        """
        return isinstance(counterpart, BasicPipingOutConnector)

    def _implement_connection(self, counterpart: BasicPipingOutConnector) -> None:
        """
        Implements the connection with a counterpart connector.

        Parameters
        ----------
        counterpart : BasicPipingOutConnector
            The counterpart connector to connect with.

        Raises
        ------
        TypeError
            If `counterpart` is not an instance of BasicPipingOutConnector.
        """
        if not isinstance(counterpart, BasicPipingOutConnector):
            raise TypeError("Countertype needs to be a BasicPipingOutConnector")
        counterpart._implement_connection(self)


class BasicPipingOutConnector(Connector):
    """Extends the Connector for a basic piping out connection in DEXPI. For
    this, it wraps a PipingTargetItem.

    Inherited Attributes
    -------------------
    label : str
        The label for the connector, inherited from `Connector`.
    kwinfos : dict, optional
        Additional keyword arguments for the connector, inherited from
        `Connector`.

    Attributes
    ----------
    piping_network_segment : piping.PipingNetworkSegment
        The PipingNetworkSegment wrapped by the connection.

    """

    def __init__(
        self,
        label: str,
        piping_network_segment: piping.PipingNetworkSegment,
        kwinfos: dict = None,
    ) -> None:
        """
        Initializes a BasicPipingOutConnector.

        Parameters
        ----------
        label : str
            The label for the connector.
        piping_network_segment : piping.PipingNetworkSegment
            The PipingNetworkSegment associated with the connection.
        kwinfos : dict, optional
            Additional keyword arguments for the connector.

        Raises
        ------
        TypeError
            If `piping_network_segment` is not an instance of PipingNetworkSegment.
        """
        if not isinstance(piping_network_segment, piping.PipingNetworkSegment):
            raise TypeError("Reference object must be a DEPXI PipingNetworkSegment")
        super().__init__(label, kwinfos)
        self.piping_network_segment = piping_network_segment

    def assess_valid_counterpart(self, counterpart: Connector) -> bool:
        """Implement abstract method to check for counterpart compatibility.
        Counterpart is valid if it is a BasicPipingInConnector

        Parameters
        ----------
        counterpart : Connector
            The candidate counterpart to be assessed for connection
            compatibility.

        Returns
        -------
        bool
            True if counterpart is suitable for connection, False otherwise
        """
        return isinstance(counterpart, BasicPipingInConnector)

    def _implement_connection(self, counterpart: BasicPipingInConnector) -> None:
        """
        Implements the connection with a counterpart connector, using the
        pydexpi piping toolkit.

        Parameters
        ----------
        counterpart : BasicPipingInConnector
            The counterpart connector to connect with.

        Raises
        ------
        TypeError
            If `counterpart` is not an instance of BasicPipingInConnector.
        """
        if not isinstance(counterpart, BasicPipingInConnector):
            raise TypeError("Countertype needs to be a BasicPipingInConnector")
        pt.connect_piping_network_segment(
            self.piping_network_segment,
            counterpart.target_item,
            connector_node_index=counterpart.target_node_index,
        )


class DexpiPattern(Pattern):
    """
    Extends the Pattern class to represent a DEXPI pattern. This class wrapps a
    DEXPI model and extends functionality for connecting patterns and
    incorporating other DEXPI patterns.

    Inherited Attributes
    ----------
    label : str
        The label for the pattern.
    connectors : list[Connector]
        A list of connectors associated with this pattern.
    observer_patterns : dict[str, Pattern], optional
        A dictionary of observer Pattern objects. If None, no observer
        patterns are used.
    kwinfos : dict, optional
        Additional keyword information for the pattern.

    Attributes
    ----------
    dexpi_model : DexpiModel
        The DEXPI model associated with this pattern.
    """

    def __init__(
        self,
        label: str,
        connectors: dict[str, Connector],
        dexpi_model: DexpiModel,
        observer_patterns: dict[str, Pattern] = None,
        kwinfos: dict = None,
    ) -> None:
        """
        Initializes a DexpiPattern with the given label, connectors, and DEXPI
        model.

        Parameters
        ----------
        label : str
            The label for the pattern.
        connectors : dict[str, Connector]
            A list of connectors associated with this pattern.
        dexpi_model : DexpiModel
            The DEXPI model associated with this pattern.
        observer_patterns : dict[str, Pattern], optional
            A dictionary of observer Pattern objects. If None, no observer
            patterns are used.
        kwinfos : dict, optional
            Additional keyword information for the pattern.

        Raises
        ------
        TypeError
            If any of the connectors are not valid connector subtypes.
        """
        valid_connector_types = (
            BasicPipingInConnector,
            BasicPipingOutConnector,
        )
        # For backwards compatibility, cast a list of connectors to a dict
        if isinstance(connectors, list):
            connectors = {connector.label: connector for connector in connectors}

        for connector in connectors.values():
            if not isinstance(connector, valid_connector_types):
                raise TypeError(
                    "A connector in the passed connectors does nothave the correct type."
                )

        super().__init__(label, connectors, observer_patterns, kwinfos)
        self.dexpi_model = dexpi_model

    def _implement_incorporation(self, counterpart: DexpiPattern) -> None:
        """
        Incorporates the specified counterpart DexpiPattern into this pattern,
        using the pydexpi model toolkit.

        Parameters
        ----------
        counterpart : DexpiPattern
            The counterpart pattern to be incorporated.

        Raises
        ------
        TypeError
            If `counterpart` is not an instance of DexpiPattern.
        """
        if not isinstance(counterpart, DexpiPattern):
            raise TypeError("Counterpart must be a DexpiPattern")
        super()._implement_incorporation(counterpart)
        mt.import_model_contents_into_model(self.dexpi_model, [counterpart.dexpi_model])



"""Tests a simple, dummy generator setup with dummy patterns and a
RandomGeneratorFunction."""

import random

import pytest

from pydexpi.syndata.generation_history import GenerationHistory
from pydexpi.syndata.generator import SyntheticPIDGenerator
from pydexpi.syndata.generator_function import CappingFunction, RandomGeneratorFunction
from pydexpi.syndata.pattern import Pattern


@pytest.fixture
def test_generator_factory(simple_distribution_factory):
    """Fixture to create a test generator."""

    def make_a_test_generator(with_capping: bool = False, random_seed: int | None = 42):
        """Create a test generator.

        Parameters
        ----------
        with_capping : bool, optional
            Whether to include a capping function, by default False
        random_seed : int, optional
            Seed for random number generation, by default 42
        """
        # Set random seed if provided for reproducible tests
        if random_seed is not None:
            random.seed(random_seed)

        distributions = {name: simple_distribution_factory(name) for name in ["Distr1", "Distr2"]}
        the_function = RandomGeneratorFunction(distributions)
        capping_function = CappingFunction() if with_capping else None
        the_generator = SyntheticPIDGenerator(
            the_function, capping_function=capping_function, max_steps=5
        )
        return the_generator

    return make_a_test_generator


def test_generator_constructor(test_generator_factory):
    """Test creating a generator."""
    generator = test_generator_factory(with_capping=False)
    assert isinstance(generator, SyntheticPIDGenerator)


def test_generate_pattern(test_generator_factory):
    """Test generating a pattern with the synthetic PID generator"""
    generator = test_generator_factory(with_capping=False)
    pattern = generator.generate_pattern("New label")
    assert isinstance(pattern, Pattern)
    assert generator._current_step > 0
    assert pattern.label == "New label"


def test_reset(test_generator_factory):
    """Test resetting the generator."""
    generator = test_generator_factory(with_capping=False)
    _ = generator.generate_pattern("New label")
    generator.reset()
    assert generator._current_step == 0


def test_generate_pattern_with_capping(test_generator_factory):
    """Test generating a pattern with capping enabled."""
    generator = test_generator_factory(with_capping=True)
    assert generator.capping_function is not None

    pattern = generator.generate_pattern("Capped Pattern")
    assert isinstance(pattern, Pattern)

    # Capping removes all connectors
    assert len(pattern.connectors) == 0


def test_generation_history_initialization(test_generator_factory):
    """Test that the generation history is properly initialized."""
    generator = test_generator_factory(with_capping=False)
    history = generator.get_generation_history()

    assert isinstance(history, GenerationHistory)
    assert len(history.history) == 0


def test_generation_history_recording(test_generator_factory):
    """Test that steps are recorded in the generation history."""
    generator = test_generator_factory(with_capping=True)

    # Generate a pattern which should create history entries
    generator.generate_pattern("Test Pattern")

    # Get the history and verify it has entries
    history = generator.get_generation_history()
    assert len(history.history) > 0

    # The first step should be the initialization step
    first_step = history.history[0]
    assert "generator_step_type" in first_step.keys()
    assert first_step["generator_step_type"] == "initialization"

    # Last step should be a capping step
    last_step = history.history[-1]
    assert last_step["generator_step_type"] == "capping"


def test_generation_history_reset(test_generator_factory):
    """Test that the generation history is reset when the generator is reset."""
    generator = test_generator_factory(with_capping=False)

    # Generate a pattern to create history
    generator.generate_pattern("First Pattern")
    assert len(generator.get_generation_history().history) > 0

    # Reset the generator
    generator.reset()

    # History should be empty after reset
    assert len(generator.get_generation_history().history) == 0

    # Generate a new pattern and verify history is populated again
    generator.generate_pattern("Second Pattern")
    assert len(generator.get_generation_history().history) > 0



"""Utility functions for parsing and error handling in the inbuilt Loader Modules."""

from functools import wraps

from pydantic import ValidationError

from pydexpi.dexpi_classes.pydantic_classes import DexpiBaseModel
from pydexpi.loaders.proteus_serializer.core import ErrorLevels, ParserModule


class ErrorTemplates:
    """Class to hold error message templates for the ProteusLoader."""

    @staticmethod
    def id_not_found(element_name: str) -> tuple[str, ErrorLevels]:
        """Returns an error message template for when an ID is not found."""
        return f"ID not found for element '{element_name}'. Element Skipped.", ErrorLevels.ERROR

    @staticmethod
    def skip_pass(pass_name: str, element_name: str) -> tuple[str, ErrorLevels]:
        """Returns an error message template for skipping a pass."""
        return (
            f"'{pass_name}' pass skipped for element '{element_name}' due to error in previous "
            f"pass. Element Skipped.",
            ErrorLevels.INFO,
        )

    @staticmethod
    def assoc_adding_error(element_name: str, associated_obj_id: str) -> tuple[str, ErrorLevels]:
        """Returns an error message template for when an association cannot be added."""
        return (
            f"Object {associated_obj_id} could not be added as reference to {element_name}.",
            ErrorLevels.ERROR,
        )

    @staticmethod
    def no_assoc_ctrl(element_name: str, associated_obj_id: str) -> tuple[str, ErrorLevels]:
        """Returns an error message template for when an association is missing in control pass."""
        return (
            f"Association of {element_name} with {associated_obj_id} is missing in control pass. "
            f"Attempt adding association.",
            ErrorLevels.WARNING,
        )

    @staticmethod
    def assoc_added_ctrl(
        element_name: str, associated_obj_id: str, field_name: str, reverse: bool = False
    ) -> tuple[str, ErrorLevels]:
        """Returns an error message template for when an association is added in control pass."""
        if reverse:
            messg = (
                f"Association of {element_name} with {associated_obj_id} added in control pass in "
                f"reverse order to field {field_name}.",
                ErrorLevels.INFO,
            )
        else:
            messg = (
                f"Association of {element_name} with {associated_obj_id} added in control pass to "
                f"field {field_name}.",
                ErrorLevels.INFO,
            )
        return messg

    @staticmethod
    def assoc_not_added_ctrl(element_name: str, associated_obj_id: str) -> tuple[str, ErrorLevels]:
        """Returns error message template for when an association is not added in control pass."""
        return (
            f"Association of {element_name} with {associated_obj_id} could not be added in control "
            f"pass.",
            ErrorLevels.ERROR,
        )

    @staticmethod
    def inval_assoc_type(element_name: str, assoc_type: str) -> tuple[str, ErrorLevels]:
        """Returns an error message template for when an association type is invalid."""
        return (
            f"Invalid association type '{assoc_type}' for {element_name}. Association skipped.",
            ErrorLevels.ERROR,
        )


def add_by_inferring_type(to_be_added: DexpiBaseModel, to_be_added_to: DexpiBaseModel):
    """Adds an attribute to a class by inferring the correct type.

    Tries to set all attributes or add them to existing lists in that field. Returns the name of
    the field found. Ignores suitable fields that are already set to a non-None value.

    Parameters
    ----------
    to_be_added : DexpiBaseModel
        The object to be added.
    to_be_added_to : DexpiBaseModel
        The object to which the `to_be_added` object should be added.

    Returns
    -------
    str
        The name of the field in `to_be_added_to` where `to_be_added`
        was successfully added.

    Raises
    ------
    ValueError
        If no suitable, None field is found in `to_be_added_to` for `to_be_added`.
    """
    field_found = False
    found_field = None
    for field in to_be_added_to.__class__.model_fields:
        try:
            curr_value = getattr(to_be_added_to, field)
            if curr_value is None:
                setattr(to_be_added_to, field, to_be_added)
                field_found = True
                found_field = field
                break

        except (ValidationError, AttributeError):
            pass
        try:
            new_list = list(getattr(to_be_added_to, field))
            new_list.append(to_be_added)
            setattr(to_be_added_to, field, new_list)
            field_found = True
            found_field = field
            break

        except (ValidationError, TypeError):
            pass

    if not field_found:
        raise ValueError(f"No suitable field found in {to_be_added_to} for {to_be_added}")

    return found_field


def is_associated_with(member: DexpiBaseModel, container: DexpiBaseModel) -> bool:
    """Checks if the object is already associated with the given object.

    Parameters
    ----------
    member : DexpiBaseModel
        The object to check for association.
    container : DexpiBaseModel
        The object to check against for association.

    Returns
    -------
    bool
        True if the member is associated with the container, False otherwise.
    """
    for field in container.__class__.model_fields:
        if isinstance(getattr(container, field), list):
            if member in getattr(container, field):
                return True
        elif getattr(container, field) == member:
            return True
    return False


def filter_none(items: list) -> list:
    """Filters out None values from a list.

    Parameters
    ----------
    items : list
        The list of items to filter.

    Returns
    -------
    list
        A new list containing only non-None items.
    """
    return [item for item in items if item is not None]


def redirect_errors_to_registry(func):
    """Create a decorator to redirect exceptions to the error registry and return None."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if not args:
                raise ValueError(
                    "The first argument 'self' must be an instance of LoaderModule. "
                    "Decorator can only be used on methods of LoaderModule"
                )
            if not isinstance(args[0], ParserModule):
                raise ValueError(
                    "The first argument 'self' must be an instance of LoaderModule. "
                    "Decorator can only be used on methods of LoaderModule"
                )
            self = args[0]
            message = (
                f"Error in parsing step {func.__name__} in {self.__class__.__name__}: {str(e)}"
            )
            self.register_error(
                message=message,
                level=ErrorLevels.ERROR,
                exception=e,
            )
            return None

    return wrapper
